@@ -1,12 +1,11 @@
 /*
     Title:      Poly/ML Console Window.
 
-    Copyright (c) 2000 David C. J. Matthews
+    Copyright (c) 2000, 2015 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -68,6 +67,7 @@
 #include "console.h"
 #include "../polyexports.h"
 #include "processes.h"
+#include "polystring.h"
 
 /*
 This module takes the place of the Windows console which
@@ -87,7 +87,6 @@ DCJM 30/5/2000.
 #endif
 
 HWND hMainWindow = NULL; // Main window - exported.
-bool useConsole;         // False if callers should read from stdin.
 HINSTANCE hApplicationInstance;     // Application instance (exported)
 static HANDLE  hReadFromML; // Handles to pipe from ML thread
 static WNDPROC  wpOrigEditProc; // Saved window proc.
@@ -101,12 +100,11 @@ static int  nReadPosn;      // Position of last read (<= nAvailable)
 static CRITICAL_SECTION csIOInterlock;
 HANDLE hInputEvent;  // Signalled when input is available.
 static HWND hDDEWindow;     // Window to handle DDE requests from ML thread.
+HANDLE hOldStdin = INVALID_HANDLE_VALUE;
 
-static char *lpszServiceName;
-
-static LPTSTR   lpArgs[100]; // Argument list.
-static int      nArgs;
-static int initDDEControl(const char *lpszName);
+static LPTSTR*  lpArgs = 0; // Argument list.
+static int      nArgs = 0;
+static int initDDEControl(const TCHAR *lpszName);
 static void uninitDDEControl(void);
 static DWORD dwDDEInstance;
 
@@ -114,7 +112,13 @@ static int nInitialShow; // Value of nCmdShow passed in.
 static bool isActive = false;
 
 // Default DDE service name.
-#define POLYMLSERVICE   "PolyML"
+#define POLYMLSERVICE   _T("PolyML")
+
+#ifdef UNICODE
+#define DDECODEPAGE CP_WINUNICODE
+#else
+#define DDECODEPAGE CP_WINANSI
+#endif
 
 
 /* Messages interpreted by the main window thread. */
@@ -122,6 +126,7 @@ static bool isActive = false;
 #define WM_DDESTART     (WM_APP+1)
 #define WM_DDESTOP      (WM_APP+2)
 #define WM_DDEEXEC      (WM_APP+3)
+#define WM_DDESERVINIT      (WM_APP+4)
 
 /* These functions are called by the I/O routines to test for input and
    to read from the keyboard. */
@@ -281,7 +286,7 @@ static void CheckForBufferSpace(int nChars)
    DDE calls and processed the message list in an "interrupt" routine.
    That complicates the Windows interface so now the ML thread sends messages
    to the main window thread to do the work. */
-HCONV StartDDEConversation(char *serviceName, char *topicName)
+HCONV StartDDEConversation(TCHAR *serviceName, TCHAR *topicName)
 {
     return (HCONV)SendMessage(hDDEWindow, WM_DDESTART, (WPARAM)serviceName, (LPARAM)topicName);
 }
@@ -296,21 +301,29 @@ LRESULT ExecuteDDE(char *command, HCONV hConv)
     return SendMessage(hDDEWindow, WM_DDEEXEC, (WPARAM)hConv, (LPARAM)command);
 }
 
+// This is called by the main Poly/ML thread after the arguments have been processed.
+void SetupDDEHandler(const TCHAR *lpszServiceName)
+{
+    SendMessage(hDDEWindow, WM_DDESERVINIT, 0, (LPARAM)lpszServiceName);
+}
 
 LRESULT CALLBACK DDEWndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
 {
     switch (uMsg) 
     {
+    case WM_DDESERVINIT:
+        return initDDEControl((const TCHAR*)lParam);
+
     case WM_DDESTART:
         {
             HCONV hcDDEConv;
             HSZ hszServiceName, hszTopicName;
-            char *serviceName = (char*)wParam;
-            char *topicName = (char*)lParam;
+            TCHAR *serviceName = (TCHAR*)wParam;
+            TCHAR *topicName = (TCHAR*)lParam;
             hszServiceName =
-                DdeCreateStringHandle(dwDDEInstance, serviceName, CP_WINANSI);
+                DdeCreateStringHandle(dwDDEInstance, serviceName, DDECODEPAGE);
             hszTopicName =
-                DdeCreateStringHandle(dwDDEInstance, topicName, CP_WINANSI);
+                DdeCreateStringHandle(dwDDEInstance, topicName, DDECODEPAGE);
             hcDDEConv =
                 DdeConnect(dwDDEInstance, hszServiceName, hszTopicName, NULL);
             DdeFreeStringHandle(dwDDEInstance, hszServiceName);
@@ -509,7 +522,11 @@ static BOOL CALLBACK AboutProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lP
     }
 }
 
+#ifdef UNICODE
+#define CF_TEXTFORMAT   CF_UNICODETEXT
+#else
 #define CF_TEXTFORMAT   CF_TEXT
+#endif
 
 /* This is the main window procedure. */
 LRESULT CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
@@ -548,10 +565,10 @@ LRESULT CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
                                     FALSE, FALSE, FALSE, ANSI_CHARSET,
                                     OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                                     DEFAULT_QUALITY, FIXED_PITCH | FF_MODERN,
-                                    "Courier");
+                                    _T("Courier"));
                 if (hFont) SendMessage(hEditWnd, WM_SETFONT, (WPARAM)hFont, 0);
  
-                SendMessage(hEditWnd, WM_SETTEXT, 0, (LPARAM) "");
+                SetWindowText(hEditWnd, _T(""));
                 return 0; /* Succeeded */
             }
  
@@ -650,7 +667,6 @@ LRESULT CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
                 processes->Exit(0);
             return 0;
 
-
         case WM_ADDTEXT:
             // Request from the input thread to add some text.
             {
@@ -690,7 +706,17 @@ static DWORD WINAPI InThrdProc(LPVOID lpParameter)
             return 0;
         buff[dwRead] = 0;
         if (! isActive) { ShowWindow(hMainWindow, nInitialShow); isActive = true; }
+#ifdef UNICODE
+        // We need to write Unicode here.  Convert it using the current code-page.
+        int wlen = MultiByteToWideChar(codePage, 0, buff, -1, NULL, 0);
+        if (wlen == 0) continue;
+        WCHAR *wBuff = new WCHAR[wlen];
+        wlen = MultiByteToWideChar(codePage, 0, buff, -1, wBuff, wlen);
+        SendMessage(hMainWindow, WM_ADDTEXT, 0, (LPARAM)wBuff);
+        delete[] wBuff;
+#else
         SendMessage(hMainWindow, WM_ADDTEXT, 0, (LPARAM)buff);
+#endif
     }
 }
 
@@ -702,19 +728,31 @@ static DWORD WINAPI MainThrdProc(LPVOID lpParameter)
     return polymain(nArgs, lpArgs, exports);
 }
 
+// Called with various control events if the input stream is a console.
+static BOOL WINAPI CtrlHandler(DWORD dwCtrlType)
+{
+    if (dwCtrlType == CTRL_C_EVENT)
+    {
+        RequestConsoleInterrupt();
+        return TRUE;
+    }
+    return FALSE;
+}
 
+// Main entry point.  Called from WinMain with a pointer to the ML code.
 int PolyWinMain(
   HINSTANCE hInstance,
   HINSTANCE hPrevInstance,
-  LPTSTR lpCmdLine,
+  LPSTR lpCmdLineUnused,
   int nCmdShow,
   exportDescription *exports
 )
 {
     HANDLE hWriteToScreen = INVALID_HANDLE_VALUE;
-    BOOL fNext = FALSE;
     DWORD dwInId, dwRes;
 
+    SetErrorMode(0); // Force a proper error report
+
     InitializeCriticalSection(&csIOInterlock);
     hInputEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
     hApplicationInstance = hInstance;
@@ -724,8 +762,59 @@ int PolyWinMain(
     // to connect it.  We use _get_osfhandle here because that
     // checks for handles passed in in the STARTUPINFO as well as
     // those inherited as standard handles.
-    if (_get_osfhandle(fileno(stdin)) == -1 ||
-        _get_osfhandle(fileno(stdout)) == -1)
+    HANDLE hStdInHandle = (HANDLE)_get_osfhandle(fileno(stdin));
+    HANDLE hStdOutHandle = (HANDLE)_get_osfhandle(fileno(stdout));
+    HANDLE hStdErrHandle = (HANDLE)_get_osfhandle(fileno(stderr));
+
+    // Do we have stdin?  If we do we need to create a pipe to buffer
+    // the input.
+    if (hStdInHandle != INVALID_HANDLE_VALUE)
+    {
+        // We're using the stdin passed in by the caller.  This may well
+        // be a pipe and in order to get reasonable performance we need
+        // to interpose a thread.  This is the only way to be able to have
+        // something we can pass to MsgWaitForMultipleObjects, in this case
+        // hInputEvent, which will indicate the input is available.
+        // Duplicate the handle because we're going to close this.
+        if (! DuplicateHandle(GetCurrentProcess(), hStdInHandle,
+                              GetCurrentProcess(), &hOldStdin, 0, TRUE, // inheritable
+                              DUPLICATE_SAME_ACCESS ))
+            return 1;
+
+        HANDLE hNewStdin = CreateCopyPipe(hOldStdin, hInputEvent);
+        if (hNewStdin == NULL) return 1;
+
+        SetConsoleCtrlHandler(CtrlHandler, TRUE); // May fail if there's no console.
+
+        // Replace the current stdin with the output from the pipe.
+        fclose(stdin);
+        int newstdin = _open_osfhandle ((INT_PTR)hNewStdin, _O_RDONLY | _O_TEXT);
+        if (newstdin != 0) _dup2(newstdin, 0);
+        fdopen(0, "rt");
+    }
+    else
+    {
+        // No stdin.  Open it on NUL.  If we actually create our own console
+        // we won't actually use this and instead we'll read from the console.
+        // In that case we won't use stdin but something else might.
+        fclose(stdin);
+        int newstdin = open("NUL", _O_RDONLY);
+        _dup2(newstdin, 0);
+        // Open it for stdio as well.  Because the entries in the FILE table
+        // are opened in order we need to do this to ensure that stdout and
+        // stderr point to the correct entries.
+        _fdopen(0, "rt");
+        hStdInHandle =  (HANDLE)_get_osfhandle(newstdin);
+        SetStdHandle(STD_INPUT_HANDLE, hStdInHandle);
+
+        // If we're not going to create a console because we have a stdout
+        // we need to set this as the original stdin.
+        if (hStdOutHandle != INVALID_HANDLE_VALUE)
+            hOldStdin = hStdInHandle;
+    }
+
+    // If we don't have a standard output we use our own console.
+    if (hStdOutHandle == INVALID_HANDLE_VALUE)
     {
         WNDCLASSEX wndClass;
         ATOM atClass;
@@ -739,8 +828,7 @@ int PolyWinMain(
         HANDLE hTemp;
         // The pipe handles we have are not inheritable.  We have to
         // make hWriteToScreen an inheritable handle so that
-        // processes we fork using "system" or "_popen"
-        // (used for profiling) can write to the screen.
+        // processes we fork using "system" can write to the screen.
         if (! DuplicateHandle(GetCurrentProcess(), hWriteToScreen,
                              GetCurrentProcess(), &hTemp, 0, TRUE, // inheritable
                              DUPLICATE_SAME_ACCESS )) {
@@ -749,34 +837,33 @@ int PolyWinMain(
         CloseHandle(hWriteToScreen);
         hWriteToScreen = hTemp;
 
-        // We never use stdin internally if we have our own console
-        // but _pipe, (used in profiling) at least, needs stdin to
-        // be non-empty.  Open it on NUL.
-        fclose(stdin);
-        int newstdin = open("NUL", _O_RDONLY);
-        _dup2(newstdin, 0);
-        // Open it for stdio as well.  Because the entries in the FILE table
-        // are opened in order we need to do this to ensure that stdout and
-        // stderr point to the correct entries.
-        fdopen(0, "rt");
-        SetStdHandle(STD_INPUT_HANDLE, (HANDLE)_get_osfhandle(newstdin));
         // Replace the standard Windows handles.
         SetStdHandle(STD_OUTPUT_HANDLE, hWriteToScreen);
-        SetStdHandle(STD_ERROR_HANDLE, hWriteToScreen);
         // Close the stdio streams.  They may have been opened
         // on dummy handles.
         fclose(stdout);
-        fclose(stderr);
         // Set up the new handles.
         int newstdout = _open_osfhandle ((INT_PTR)hWriteToScreen, _O_TEXT);
+        // We need this to be stream 1.  basicio.cpp uses this for TextIO.stdOut
         if (newstdout != 1) _dup2(newstdout, 1);
-        _dup2(newstdout, 2); // Stderr
-        // Open for stdio.
-        fdopen(1, "wt"); // == stdout
-        fdopen(2, "wt"); // == stderr
-        // Set stderr to unbuffered so that messages get written correctly.
-        // (stdout is explicitly flushed).
-        setvbuf(stderr, NULL, _IONBF, 0);
+        // A few RTS modules use stdio for output, primarily objsize and diagnostics.
+        // Previously this next line was sufficient to reopen stdout but that no longer
+        // works in VS 2015.  We have to use polyStdout now.
+        extern FILE *polyStdout;
+        polyStdout = _fdopen(1, "wt"); // == stdout
+
+        if (hStdErrHandle == INVALID_HANDLE_VALUE)
+        {
+            // If we didn't have stderr write any stderr output to our console.
+            SetStdHandle(STD_ERROR_HANDLE, hWriteToScreen);
+            fclose(stderr);
+            _dup2(newstdout, 2); // Stderr
+            extern FILE *polyStderr;
+            polyStderr = _fdopen(2, "wt"); // == stderr
+            // Set stderr to unbuffered so that messages get written correctly.
+            // (stdout is explicitly flushed).
+            setvbuf(stderr, NULL, _IONBF, 0);
+        }
 
         // Create a thread to manage the output from ML.
         HANDLE hInThread = CreateThread(NULL, 0, InThrdProc, 0, 0, &dwInId);
@@ -825,68 +912,55 @@ int PolyWinMain(
         // read from or write to the main window.  That way if we are
         // actually using another window this will never get displayed.
         nInitialShow = nCmdShow;
-        useConsole = true;
     }
-    else {
-        // We're using the stdin passed in by the caller.  This may well
-        // be a pipe and in order to get reasonable performance we need
-        // to interpose a thread.  This is the only way to be able to have
-        // something we can pass to MsgWaitForMultipleObjects, in this case
-        // hInputEvent, which will indicate the input is available.
-        HANDLE hOldStdin;
-        // Duplicate the handle because we're going to close this.
-        if (! DuplicateHandle(GetCurrentProcess(), GetStdHandle(STD_INPUT_HANDLE),
-                              GetCurrentProcess(), &hOldStdin, 0, TRUE, // inheritable
-                              DUPLICATE_SAME_ACCESS ))
-            return 1;
-
-        HANDLE hNewStdin = CreateCopyPipe(hOldStdin, hInputEvent);
-        if (hNewStdin == NULL) return 1;
-
-        // Replace the current stdin with the output from the pipe.
-        fclose(stdin);
-        int newstdin = _open_osfhandle ((INT_PTR)hNewStdin, _O_RDONLY | _O_TEXT);
-        if (newstdin != 0) _dup2(newstdin, 0);
-        fdopen(0, "rt");
-        useConsole = false;
-    }
-
-    // Convert the command line into Unix-style arguments.
-    LPTSTR lpCommandLine = GetCommandLine();
-    nArgs = 0;
-    while (*lpCommandLine != 0)
+    // We had a stdout but maybe not stderr.  We could choose to direct stderr output
+    // to the provided stdout but maybe that's not what the user wants.  Instead
+    // we open one on NUL.
+    else if (hStdErrHandle == INVALID_HANDLE_VALUE)
     {
-        if (*lpCommandLine == '"')
-        {
-            // Treat quoted items as a whole.
-            lpCommandLine++;
-            lpArgs[nArgs++] = lpCommandLine;
-            while (*lpCommandLine != 0 && *lpCommandLine != '"')
-                lpCommandLine++;
-            if (*lpCommandLine != 0) *lpCommandLine++ = 0;
-        }
-        else
-        {
-            // Not quoted - look for a separating space.
-            lpArgs[nArgs++] = lpCommandLine;
-            while (*lpCommandLine != 0 && *lpCommandLine != ' ')
-                lpCommandLine++;
-        }
-        // Remove multiple spaces.
-        while (*lpCommandLine == ' ') *lpCommandLine++ = 0;
+        fclose(stderr);
+        int newstderr = open("NUL", _O_WRONLY);
+        _dup2(newstderr, 2); // Stderr
+        _fdopen(2, "wt"); // == stderr
+        SetStdHandle(STD_ERROR_HANDLE, (HANDLE)_get_osfhandle(newstderr));
+        // Set stderr to unbuffered so that messages get written correctly.
+        // (stdout is explicitly flushed).
+        setvbuf(stderr, NULL, _IONBF, 0);
+    }
 
-        if (nArgs == sizeof(lpArgs)/sizeof(lpArgs[0])) break;
-
-        // Extract the service name argument.
-        if (strcmp(lpArgs[nArgs-1], "-pServiceName") == 0)
+    // Set nArgs and lpArgs to the command line arguments.
+    // Convert the command line into Unix-style arguments.  There isn't a
+    // CommandLineToArgvA function so we have to use the Unicode version and
+    // convert the results.
+    {
+        // Get the unicode args
+        LPWSTR *uniArgs = CommandLineToArgvW(GetCommandLineW(), &nArgs);
+#ifdef UNICODE
+        lpArgs = uniArgs;
+#else
+        if (uniArgs != NULL)
         {
-            nArgs--;
-            fNext = TRUE;
-        }
-        else {
-            if (fNext) lpszServiceName = lpArgs[--nArgs];
-            fNext = FALSE;
+            lpArgs = (LPSTR*)calloc(nArgs, sizeof(LPSTR));
+            if (lpArgs != 0)
+            {
+                for (int i = 0; i < nArgs; i++)
+                {
+                    // See how much space will be needed
+                    int space =
+                        WideCharToMultiByte(CP_ACP, 0, uniArgs[i], -1, NULL, 0, NULL, NULL);
+                    if (space == 0) break; // Failed for some reason
+                    // Allocate the space then do the conversion
+                    LPSTR buff = (LPSTR)malloc(space);
+                    if (buff == 0) break;
+                    int result =
+                        WideCharToMultiByte(CP_ACP, 0, uniArgs[i], -1, buff, space, NULL, NULL);
+                    if (result == 0) { free(buff); break; }
+                    lpArgs[i] = buff;
+                }
+            }
+            LocalFree(uniArgs);
         }
+#endif
     }
 
     // Create an internal hidden window to handle DDE requests from the ML thread.
@@ -897,13 +971,13 @@ int PolyWinMain(
         wndClass.cbSize = sizeof(wndClass);
         wndClass.lpfnWndProc = DDEWndProc; 
         wndClass.hInstance = hInstance; 
-        wndClass.lpszClassName = "PolyMLDDEWindowClass";
+        wndClass.lpszClassName = _T("PolyMLDDEWindowClass");
 
         if ((atClass = RegisterClassEx(&wndClass)) == 0) return 1;
 
         hDDEWindow = CreateWindow(
             (LPTSTR)(intptr_t)atClass,
-            "Poly/ML-DDE",
+            _T("Poly/ML-DDE"),
             WS_OVERLAPPEDWINDOW,
             CW_USEDEFAULT,
             CW_USEDEFAULT,
@@ -916,8 +990,6 @@ int PolyWinMain(
             );
     }
 
-    initDDEControl(lpszServiceName);
-
     // Call the main program to do the rest of the initialisation.
     HANDLE hMainThread = CreateThread(NULL, 0, MainThrdProc, exports, 0, &dwInId);
 
@@ -960,15 +1032,15 @@ HDDEDATA CALLBACK DdeCallback(UINT uType, UINT uFmt, HCONV hconv,
             {
                 // See what the message is.  The only ones we
                 // handle are interrupt and terminate.
-                CHAR buff[256];
+                TCHAR buff[256];
                 buff[0] = 0;
                 DdeGetData(hdata, (LPBYTE)buff, sizeof(buff), 0);
-                if (lstrcmpi(buff, POLYINTERRUPT) == 0)
+                if (lstrcmpi(buff, _T(POLYINTERRUPT)) == 0)
                 {
                     RequestConsoleInterrupt();
                     return (HDDEDATA) DDE_FACK;
                 }
-                if (lstrcmpi(buff, POLYTERMINATE) == 0)
+                if (lstrcmpi(buff, _T(POLYTERMINATE)) == 0)
                 {
                     processes->Exit(0);
                     return (HDDEDATA) DDE_FACK;
@@ -981,9 +1053,9 @@ HDDEDATA CALLBACK DdeCallback(UINT uType, UINT uFmt, HCONV hconv,
     } 
 } 
 
-static int initDDEControl(const char *lpszName)
+static int initDDEControl(const TCHAR *lpszName)
 {
-    HSZ hszServiceName;
+    // Start the DDE service.  This receives remote requests.
     if (DdeInitialize(&dwDDEInstance, DdeCallback,
         APPCLASS_STANDARD | CBF_FAIL_ADVISES | CBF_FAIL_POKES |
         CBF_FAIL_REQUESTS | CBF_SKIP_ALLNOTIFICATIONS, 0)
@@ -993,7 +1065,7 @@ static int initDDEControl(const char *lpszName)
     // If we were given a service name we register that,
     // otherwise we use the default name.
     if (lpszName == 0) lpszName = POLYMLSERVICE;
-    hszServiceName = DdeCreateStringHandle(dwDDEInstance, lpszName, CP_WINANSI);
+    HSZ hszServiceName = DdeCreateStringHandle(dwDDEInstance, lpszName, DDECODEPAGE);
     if (hszServiceName == 0) return 0;
 
     DdeNameService(dwDDEInstance, hszServiceName, 0L, DNS_REGISTER);
@@ -1050,7 +1122,7 @@ static DWORD WINAPI copyThread(LPVOID lpParameter)
 }
 
 // This thread is used when the caller has provided a standard input
-// stream and we're using that and not out console.  It copies the
+// stream and we're using that and not our console.  It copies the
 // standard input to a pipe and the ML code uses that as its input.
 // This way we can set hInputEvent whenever input is available.
 void CopyPipe::threadFunction()
@@ -1068,6 +1140,13 @@ void CopyPipe::threadFunction()
 
         if (dwRead == 0) // End-of-stream
         {
+            // If we are reading from the (Windows) console and the user presses ctrl-C we
+            // may get a ERROR_OPERATION_ABORTED error.
+            if (GetLastError() == ERROR_OPERATION_ABORTED)
+            {
+                SetLastError(0); // Reset this.  We may have a normal EOF next.
+                continue;
+            }
             // Normal exit.  Indicate EOF
             SetEvent(hEvent);
             return;
