@@ -1,12 +1,11 @@
 /*
     Title:      Poly/ML Console Window.
 
-    Copyright (c) 2000 David C. J. Matthews
+    Copyright (c) 2000, 2015 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -34,14 +33,21 @@ extern unsigned getConsoleInput(char *buff, int nChars);
 extern HANDLE CreateCopyPipe(HANDLE hInput, HANDLE hEvent);
 
 extern HWND hMainWindow; /* Handle to main window - NULL if none. */
-extern bool useConsole; /* non-zero if we should use the console for input. */
+
 extern HINSTANCE hApplicationInstance; /* Application instance */
 
 /* DDE requests. */
-extern HCONV StartDDEConversation(char *serviceName, char *topicName);
+extern HCONV StartDDEConversation(TCHAR *serviceName, TCHAR *topicName);
 extern void CloseDDEConversation(HCONV hConv);
 extern LRESULT ExecuteDDE(char *command, HCONV hConv);
 
 extern HANDLE hInputEvent; // Handle to console input event
 
+extern HANDLE hOldStdin; // If stdin has been provided
+
+// If the original stdin handle was missing we're using our console.
+#define useConsole (hOldStdin == INVALID_HANDLE_VALUE)
+
+extern void SetupDDEHandler(const TCHAR *lpszName);
+
 #endif
