@@ -27,6 +27,10 @@
 #error "No configuration file"
 #endif
 
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#include <winsock2.h>
+#endif
+
 #ifdef HAVE_WINDOWS_H
 #include <windows.h>
 #endif
@@ -39,14 +43,6 @@
 #include <string.h>
 #endif
 
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
-#ifdef USEWINSOCK2
-#include <winsock2.h>
-#else
-#include <winsock.h>
-#endif
-#endif
-
 #include "errors.h"
 
 static struct {
