@@ -1,12 +1,11 @@
 /*
     Title:  exporter.h - Export a function as an object or C file
 
-    Copyright (c) 2006 David C.J. Matthews
+    Copyright (c) 2006, 2015 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -40,7 +39,7 @@ extern Handle exportPortable(TaskData *mdTaskData, Handle args);
 class Exporter
 {
 public:
-    Exporter();
+    Exporter(unsigned int h=0);
     virtual ~Exporter();
     virtual void exportStore(void) = 0;
 
@@ -59,11 +58,13 @@ public:
     const char *errorMessage;
 
 protected:
+    unsigned int hierarchy;
     struct _memTableEntry *memTable;
     unsigned memTableEntries;
     unsigned ioMemEntry; // The index corresponding to the IO area.
     unsigned ioSpacing;
     PolyObject *rootFunction; // Address of the root function.
+    unsigned newAreas;
 };
 
 // The object-code exporters all use a similar string table format
@@ -110,4 +111,12 @@ public:
     unsigned tombs;
 };
 
+class ClearWeakByteRef: public ScanAddress
+{
+public:
+    ClearWeakByteRef() {}
+    virtual PolyObject *ScanObjectAddress(PolyObject *base) { return base; }
+    virtual void ScanAddressesInObject(PolyObject *base, POLYUNSIGNED lengthWord);
+};
+
 #endif
