@@ -4,12 +4,11 @@
 
     Copyright (c) 2000-7
         Cambridge University Technical Services Limited
-    Further development Copyright David C.J. Matthews 2008-2011.
+    Further development Copyright David C.J. Matthews 2008-2015.
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -81,17 +80,20 @@ extern "C" {
 #include <malloc.h>
 #endif
 
-#ifdef HAVE_TCHAR_H
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
 #include <tchar.h>
 #else
 typedef char TCHAR;
 #define _T(x) x
 #endif
 
+#include "globals.h"
+// We need to include globals.h before <new> in mingw64 otherwise
+// it messes up POLYUFMT/POLYSFMT.
+
 #include <ffi.h>
 #include <new>
 
-#include "globals.h"
 #include "arb.h"
 #include "reals.h"
 #include "foreign.h"
@@ -251,6 +253,10 @@ typedef struct {
 
 static Volatile *vols;
 static PLock volLock; // Mutex to protect vols.
+// TODO: There is a theoretical risk of deadlock if any ML allocation is made while this
+// lock is held.  An allocation can result in a GC which requires all threads to release
+// ML memory but another thread could block waiting for the mutex.
+// N.B. raising an exception involves an allocation.
 
 #define FIRST_VOL 0
 
@@ -268,22 +274,6 @@ static struct _cbStructEntry {
 static unsigned callBackEntries = 0;
 static PLock callbackTableLock; // Mutex to protect table.
 
-
-/**********************************************************************
- *
- *  Malloc / Free Wrappers
- *   
- **********************************************************************/
-
-static POLYUNSIGNED malloc_count = 0;
-#if 0
-#define Vmalloc(where,size) {where = malloc(size); printf("malloc: %p,%d\n",where,size); fflush(stdout); malloc_count++;}
-#else
-#define Vmalloc(where,size) {where = malloc(size); malloc_count++;}
-#endif
-#define Vfree(p) { free(p);  malloc_count--;}
-
-
 /**********************************************************************
  *
  *  Volatile Allocation
@@ -349,8 +339,12 @@ static Handle vol_alloc_with_c_space (TaskData *taskData, POLYUNSIGNED size)
     PLocker plocker(&volLock);
     Handle res = vol_alloc(taskData);
     trace(("size= %" POLYUFMT "\n",size));
-    Vmalloc( C_POINTER(UNVOLHANDLE(res)), size );
+    void *p = malloc(size);
+    if (p == 0)
+        RAISE_EXN("Insufficient memory");
+    C_POINTER(UNVOLHANDLE(res)) = p;
     OWN_C_SPACE(UNVOLHANDLE(res)) = true;
+
     return res;
 }
 
@@ -425,7 +419,7 @@ static void sanity_check_vols(void)
    store so we need to call address(alloc s). */
 static Handle allocate (TaskData *taskData, Handle h)
 { TRACE; {
-    POLYUNSIGNED size = get_C_ulong(taskData, DEREFWORD(h)); /* bytes */
+    POLYUNSIGNED size = getPolyUnsigned(taskData, DEREFWORD(h)); /* bytes */
     Handle space = vol_alloc_with_c_space(taskData, size);
     return space;
 }}
@@ -685,7 +679,7 @@ void Foreign::GarbageCollect(ScanAddress *process)
             {
                 // Can now free this.            
                 trace(("Freeing malloc space of <%" POLYUFMT ">\n",from));
-                Vfree(vols[from].C_pointer);
+                free(vols[from].C_pointer);
                 vols[from].C_pointer = 0;
                 vols[from].Own_C_space = false;
             }
@@ -702,7 +696,6 @@ void Foreign::GarbageCollect(ScanAddress *process)
         }
     }
     next_vol = to;
-    info(("unfreed mallocs=<%" POLYUFMT "> next_vol=<%" POLYUFMT ">\n", malloc_count, next_vol));
     
     /* Callback table.  Added DCJM 12/4/04.  We always process these as strong references.
     For the time being at any rate we treat these as permanent entries so that once a
@@ -726,7 +719,12 @@ static Handle load_lib (TaskData *taskData, Handle string)
     TCHAR name[500];
     
     Poly_string_to_C(DEREFWORD(string), name, sizeof(name)/sizeof(TCHAR));
+
+#if (defined(_WIN32) && defined(UNICODE))
+    info(("<%S>\n", name));
+#else
     info(("<%s>\n", name));
+#endif
     // If the name is the null string use the current executable.
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
     HINSTANCE lib;
@@ -735,7 +733,11 @@ static Handle load_lib (TaskData *taskData, Handle string)
     if (lib == NULL) 
     {
         char buf[256];
+#if (defined(_WIN32) && defined(UNICODE))
+        sprintf(buf, "load_lib <%S> : %lu", name, GetLastError());
+#else
         sprintf(buf, "load_lib <%s> : %lu", name, GetLastError());
+#endif
         RAISE_EXN(buf);
     }
 
@@ -767,9 +769,9 @@ static Handle load_lib (TaskData *taskData, Handle string)
 
 static Handle load_sym (TaskData *taskData, Handle h)
 {
-    TCHAR name[500];
+    char name[500];
     
-    Poly_string_to_C(DEREFHANDLE(h)->Get(1), name, sizeof(name)/sizeof(TCHAR));
+    Poly_string_to_C(DEREFHANDLE(h)->Get(1), name, sizeof(name));
     info(("<%s>\n", name));
     
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
@@ -778,7 +780,7 @@ static Handle load_sym (TaskData *taskData, Handle h)
     if (sym == NULL) 
     {
         char buf[256];
-        sprintf(buf, "load_sym <%s> : %lu", name,GetLastError());
+        sprintf(buf, "load_sym <%s> : %lu", name, GetLastError());
         RAISE_EXN(buf);
     }
 
@@ -910,8 +912,9 @@ static Handle call_sym (TaskData *taskData, Handle symH, Handle argsH, Handle re
         RAISE_EXN("libffi error: ffi_prep_cif failed");
 
     // malloc memory for the result
-    void *result;
-    Vmalloc(result, result_type->size);
+    void *result = malloc(result_type->size);
+    if (result == 0)
+        RAISE_EXN("Insufficient memory to allocate space for result");
 
     processes->ThreadReleaseMLMemory(taskData);
     ffi_call(&cif, sym, result, arg_values);
@@ -970,7 +973,7 @@ static Handle fromCchar (TaskData *taskData, Handle h)
 {
     char c = *(char*)DEREFVOL(taskData, UNHANDLE(h));
     mes(("<%c>\n", c));
-    return SAVE(Buffer_to_Poly(taskData, &c,1));
+    return SAVE(C_string_to_Poly(taskData, &c,1));
 }
 
 
@@ -1116,10 +1119,6 @@ static Handle fromCuint (TaskData *taskData, Handle h)
  *  String Conversion
  *   
  **********************************************************************/
-
-#define PSTRING_LENGTH(pstr) \
-    (IS_INT((pstr)) ? 1 : (pstr)->length)
-
 static Handle fillCstring (TaskData *taskData, Handle h)
 { TRACE; {
     POLYUNSIGNED size;
@@ -1201,7 +1200,7 @@ static Handle fromCbytes (TaskData *taskData, Handle h)
     char* str = *(char**)DEREFVOL(taskData, DEREFHANDLE(h)->Get(0));
     size_t size = get_C_long(taskData, DEREFHANDLE(h)->Get(1));
     if (str == NULL) return SAVE(EmptyString());
-    else return SAVE(Buffer_to_Poly(taskData, str, size));
+    else return SAVE(C_string_to_Poly(taskData, str, size));
 }}
 
 
@@ -1509,10 +1508,10 @@ static void callbackEntryPt(ffi_cif *cif, void *ret, void* args[], void *data)
         try {
             taskData = processes->CreateNewTaskData(0, 0, 0, TAGGED(0));
         }
-        catch (std::bad_alloc) {
+        catch (std::bad_alloc &) {
             ::Exit("Unable to create thread data - insufficient memory");
         }
-        catch (MemoryException) {
+        catch (MemoryException &) {
             ::Exit("Unable to create thread data - insufficient memory");
         }
     }
@@ -1540,9 +1539,7 @@ static void callbackEntryPt(ffi_cif *cif, void *ret, void* args[], void *data)
 
     // Callbacks previously involved forking a new ML process.  They are
     // now handled on the caller's stack.
-    machineDependent->SetCallbackFunction(taskData, h, mlArgs);
-
-    Handle resultHandle = EnterPolyCode(taskData);
+    Handle resultHandle = taskData->EnterCallbackFunction(h, mlArgs);
 
     PolyWord resultWord = UNHANDLE(resultHandle);
     taskData->saveVec.reset(mark);
@@ -1580,6 +1577,9 @@ static Handle createCallbackFunction(TaskData *taskData, Handle triple, ffi_abi
 
     unsigned num_args = length_list(argTypeList->Word());
     ffi_type **arg_types = (ffi_type**)malloc(num_args * sizeof(ffi_type*));
+    if (arg_types == 0)
+        RAISE_EXN("Insufficient memory to allocate space for arguments");
+
     PolyWord p = argTypeList->Word();
     for (POLYUNSIGNED i=0; i<num_args; i++,p=Tail(p))
         arg_types[i] = ctypeToFfiType(taskData, Head(p));
@@ -1587,6 +1587,9 @@ static Handle createCallbackFunction(TaskData *taskData, Handle triple, ffi_abi
 
     // The cif needs to be on the heap so that it is available in the callback.
     ffi_cif *cif = (ffi_cif *)malloc(sizeof(ffi_cif));
+    if (cif == 0)
+        RAISE_EXN("Insufficient memory to allocate space for cif");
+
     if (ffi_prep_cif(cif, abi, num_args, result_type, arg_types) != FFI_OK)
         RAISE_EXN("libffi error: ffi_prep_cif failed");
 
