@@ -1,12 +1,11 @@
 /*
     Title:  heapsizing.cpp - parameters to adjust heap size
 
-    Copyright (c) Copyright David C.J. Matthews 2012
+    Copyright (c) Copyright David C.J. Matthews 2012, 2015
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -82,36 +81,13 @@ debugging.
 HeapSizeParameters gHeapSizeParameters;
 
 #ifdef HAVE_WINDOWS_H
-// Getting hard page counts in Windows is not easy.  Cygwin uses
-// GetProcessMemoryInfo to return the value in ru_majflt but this
-// is actually incorrect because it returns the soft page count not
-// the hard page count.  We use NtQuerySystemInformation to get the
-// total paging on the system on the basis that this is more useful
-// than the soft page faults for the process.
-// This is an undocumented interface and it's all a bit of a mess.
-
-typedef struct  {
-    INT64 pad1[4];
-    DWORD pad2[12];
-    DWORD pagesRead;
-    DWORD pad3[60];
-} SystemPerformanceInfo;
-
-typedef int (WINAPI *NtSystemInfo)(int, SystemPerformanceInfo *, ULONG, void*);
-static NtSystemInfo pFunctionPtr;
-
+// There's no (documented) way to get the per-process hard page
+// count in Windows.  Cygwin uses GetProcessMemoryInfo to return the
+// value in ru_majflt but this is actually incorrect because it returns
+// the soft page count not the hard page count.  We previously used the
+// undocumented NtQuerySystemInformation call.
 static long GetPaging(long)
 {
-    if (pFunctionPtr == 0)
-        pFunctionPtr = (NtSystemInfo) GetProcAddress(GetModuleHandle("Ntdll.dll"), "NtQuerySystemInformation");
-
-    if (pFunctionPtr != 0)
-    {
-        SystemPerformanceInfo pInfo;
-        int result = (*pFunctionPtr)(2/*SystemPerformanceInformation*/, &pInfo, sizeof(pInfo), 0);
-        if (result == 0)
-            return pInfo.pagesRead;
-    }
     return 0;
 }
 #else
@@ -140,7 +116,11 @@ HeapSizeParameters::HeapSizeParameters()
     sharingCostFactor = 2; // It doubles the cost
 }
 
-/* This macro must make a whole number of chunks */
+// These macros were originally in globals.h and used more generally.
+// Since only K_to_words is used now this can be greatly simplified.
+#define BITSPERWORD (sizeof(PolyWord)*8)
+#define ROUNDUP_UNITS(m,n)   (((m) + (n) - 1) / (n))
+#define ROUNDUP(m,n)   (ROUNDUP_UNITS(m,n) * (n))
 #define K_to_words(k) ROUNDUP((k) * (1024 / sizeof(PolyWord)),BITSPERWORD)
 
 // Returns physical memory size in bytes
@@ -178,14 +158,14 @@ void HeapSizeParameters::SetHeapParameters(POLYUNSIGNED minsize, POLYUNSIGNED ma
     // The default minimum is zero; in practice the live data size.
 
     // The default initial size is the minimum if that has been provided,
-    // otherwise it is a quarter of the memory or 64M if we can't calculate the memory.
+    // otherwise 8M words.  There are applications that only require a small
+    // heap and if we set the heap large to begin with we'll never do a
+    // full GC and reduce it.
     if (initialSize == 0)
     {
         if (minHeapSize != 0)
             initialSize = minHeapSize;
-        else if (memsize != 0)
-            initialSize =  memsize / 4;
-        else initialSize = 64 * 1024 * 1024;
+        else initialSize = 8 * gMem.DefaultSpaceSize();
         // But not more than the maximum
         if (initialSize > maxHeapSize) initialSize = maxHeapSize;
     }
@@ -590,7 +570,8 @@ bool HeapSizeParameters::getCostAndSize(POLYUNSIGNED &heapSize, double &cost, bo
     // We mustn't reduce the heap size too far.  If the application does a lot
     // of work with few allocations and particularly if it calls PolyML.fullGC
     // explicitly we could attempt to shrink the heap below the current live data size.
-    POLYUNSIGNED minForAllocation = gMem.CurrentHeapSize() + gMem.DefaultSpaceSize() * 2;
+    // Add 3*space size here.  We require 2* after a minor GC. Add 1 for rounding.
+    POLYUNSIGNED minForAllocation = gMem.CurrentHeapSize() + gMem.DefaultSpaceSize() * 3;
     if (minForAllocation > maxHeapSize) minForAllocation = maxHeapSize;
     if (sizeMin < minForAllocation) sizeMin = minForAllocation;
 
@@ -796,7 +777,7 @@ void HeapSizeParameters::RecordSharingData(POLYUNSIGNED recovery)
 Handle HeapSizeParameters::getGCUtime(TaskData *taskData) const
 {
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
-    return Make_arb_from_pair(taskData, ((FILETIME)totalGCUserCPU).dwHighDateTime, ((FILETIME)totalGCUserCPU).dwLowDateTime);
+    return Make_arb_from_Filetime(taskData, totalGCUserCPU);
 #else
     return Make_arb_from_pair_scaled(taskData, ((struct timeval)totalGCUserCPU).tv_sec, ((struct timeval)totalGCUserCPU).tv_usec, 1000000);
 #endif
@@ -805,7 +786,7 @@ Handle HeapSizeParameters::getGCUtime(TaskData *taskData) const
 Handle HeapSizeParameters::getGCStime(TaskData *taskData) const
 {
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
-    return Make_arb_from_pair(taskData, ((FILETIME)totalGCSystemCPU).dwHighDateTime, ((FILETIME)totalGCSystemCPU).dwLowDateTime);
+    return Make_arb_from_Filetime(taskData, totalGCSystemCPU);
 #else
     return Make_arb_from_pair_scaled(taskData, ((struct timeval)totalGCSystemCPU).tv_sec, ((struct timeval)totalGCSystemCPU).tv_usec, 1000000);
 #endif
@@ -926,76 +907,24 @@ void HeapSizing::Stop()
     gHeapSizeParameters.Final();
 }
 
-
-// Return the physical memory size.  Returns the maximum unsigned integer value if
-// it won't .
-#if defined(HAVE_WINDOWS_H)
-
-// Define this here rather than attempting to use MEMORYSTATUSEX since
-// it may not be in the include and we can't easily test.  The format
-// of MEMORYSTATUSVLM is the same.
-typedef struct _MyMemStatusEx {
-    DWORD dwLength;
-    DWORD dwMemoryLoad;
-    DWORDLONG ullTotalPhys;
-    DWORDLONG ullAvailPhys;
-    DWORDLONG ullTotalPageFile;
-    DWORDLONG ullAvailPageFile;
-    DWORDLONG ullTotalVirtual;
-    DWORDLONG ullAvailVirtual;
-    DWORDLONG ullAvailExtendedVirtual;
-} MyMemStatusEx;
-
-typedef VOID (WINAPI *GLOBALMEMSLVM)(MyMemStatusEx *);
-typedef BOOL (WINAPI *GLOBALMEMSEX)(MyMemStatusEx *);
-#endif
-
-
 static POLYUNSIGNED GetPhysicalMemorySize(void)
 {
     POLYUNSIGNED maxMem = 0-1; // Maximum unsigned value.
 #if defined(HAVE_WINDOWS_H)
     {
-        // This is more complicated than it needs to be.  GlobalMemoryStatus
-        // returns silly values if there is more than 4GB so GlobalMemoryStatusEx
-        // is preferred.  However, it is not in all the include files and may not
-        // be in kernel32.dll in pre-XP versions.  Furthermore at one point it was
-        // called GlobalMemoryStatusVlm.  The only way to do this portably is the
-        // hard way.
-        HINSTANCE hlibKernel = LoadLibrary("kernel32.dll");
-        if (hlibKernel)
+        MEMORYSTATUSEX memStatEx;
+        memset(&memStatEx, 0, sizeof(memStatEx));
+        memStatEx.dwLength = sizeof(memStatEx);
+        if (! GlobalMemoryStatusEx(&memStatEx))
+            memStatEx.ullTotalPhys = 0; // Clobber any rubbish since it says it failed.
+        if (memStatEx.ullTotalPhys) // If it's non-zero assume it succeeded
         {
-            MyMemStatusEx memStatEx;
-            memset(&memStatEx, 0, sizeof(memStatEx));
-            memStatEx.dwLength = sizeof(memStatEx);
-            GLOBALMEMSEX globalMemStatusEx =
-                (GLOBALMEMSEX)GetProcAddress(hlibKernel, "GlobalMemoryStatusEx");
-            GLOBALMEMSLVM globalMemStatusVlm =
-                (GLOBALMEMSLVM)GetProcAddress(hlibKernel, "GlobalMemoryStatusVlm");
-            if (globalMemStatusEx && ! (*globalMemStatusEx)(&memStatEx))
-                memStatEx.ullTotalPhys = 0; // Clobber any rubbish since it says it failed.
-            else if (globalMemStatusVlm)
-                // GlobalMemoryStatusVlm returns VOID so we assume it worked
-                (*globalMemStatusVlm) (&memStatEx);
-            FreeLibrary(hlibKernel);
-            if (memStatEx.ullTotalPhys) // If it's non-zero assume it succeeded
-            {
-                DWORDLONG dwlMax = maxMem;
-                if (memStatEx.ullTotalPhys > dwlMax)
-                    return maxMem;
-                else
-                    return (POLYUNSIGNED)memStatEx.ullTotalPhys;
-           }
+            DWORDLONG dwlMax = maxMem;
+            if (memStatEx.ullTotalPhys > dwlMax)
+                return maxMem;
+            else
+                return (POLYUNSIGNED)memStatEx.ullTotalPhys;
         }
-        // Fallback if that fails.
-
-        MEMORYSTATUS memStatus;
-        memset(&memStatus, 0, sizeof(memStatus));
-        GlobalMemoryStatus(&memStatus);
-        if (memStatus.dwTotalPhys > maxMem)
-            return maxMem;
-        else
-            return (POLYUNSIGNED)memStatus.dwTotalPhys;
     }
 
 #endif
