@@ -5,8 +5,7 @@
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -101,19 +100,14 @@
 typedef int socklen_t;
 #endif
 
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#include <winsock2.h>
+#endif
+
 #ifdef HAVE_WINDOWS_H
 #include <windows.h>
 #endif
 
-
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
-#ifdef USEWINSOCK2
-#include <winsock2.h>
-#else
-#include <winsock.h>
-#endif
-#endif
-
 #include "globals.h"
 #include "gc.h"
 #include "arb.h"
@@ -291,8 +285,6 @@ struct sk_tab_struct {
 static Handle makeHostEntry(TaskData *taskData, struct hostent *host);
 static Handle makeProtoEntry(TaskData *taskData, struct protoent *proto);
 static Handle makeServEntry(TaskData *taskData, struct servent *proto);
-static Handle makeList(TaskData *taskData, int count, char *p, int size, void *arg,
-                       Handle (mkEntry)(TaskData *, void*, char*));
 static Handle mkAftab(TaskData *taskData, void*, char *p);
 static Handle mkSktab(TaskData *taskData, void*, char *p);
 static Handle setSocketOption(TaskData *taskData, Handle args, int level, int opt);
@@ -355,36 +347,67 @@ static int GetError()
 #define MAPERROR(x) (x)
 #endif
 
-class WaitNet: public Waiter {
+
+// Wait until "select" returns.  In Windows this is used only for networking.
+class WaitSelect: public Waiter
+{
 public:
-    WaitNet(SOCKET sock, bool isOOB = false) : m_sock(sock), m_isOOB(isOOB) {}
-    void Wait(unsigned maxMillisecs);
+    WaitSelect();
+    virtual void Wait(unsigned maxMillisecs);
+    void SetRead(SOCKET fd) {  FD_SET(fd, &readSet); }
+    void SetWrite(SOCKET fd) {  FD_SET(fd, &writeSet); }
+    void SetExcept(SOCKET fd)  {  FD_SET(fd, &exceptSet); }
+    // Save the result of the select call and any associated error
+    int SelectResult(void) { return selectResult; }
+    int SelectError(void) { return errorResult; }
 private:
-    SOCKET m_sock;
-    bool m_isOOB;
+    fd_set readSet, writeSet, exceptSet;
+    int selectResult;
+    int errorResult;
+};
+
+WaitSelect::WaitSelect()
+{
+    FD_ZERO(&readSet);
+    FD_ZERO(&writeSet);
+    FD_ZERO(&exceptSet);
+    selectResult = 0;
+    errorResult = 0;
+}
+
+void WaitSelect::Wait(unsigned maxMillisecs)
+{
+    struct timeval toWait = { 0, 0 };
+    toWait.tv_sec = maxMillisecs / 1000;
+    toWait.tv_usec = (maxMillisecs % 1000) * 1000;
+    selectResult = select(FD_SETSIZE, &readSet, &writeSet, &exceptSet, &toWait);
+    if (selectResult < 0) errorResult = GETERROR;
+}
+
+class WaitNet: public WaitSelect {
+public:
+    WaitNet(SOCKET sock, bool isOOB = false);
 };
 
 // Use "select" in both Windows and Unix.  In Windows that means we
 // don't watch hWakeupEvent but that's only a hint.
-void WaitNet::Wait(unsigned maxMillisecs)
+WaitNet::WaitNet(SOCKET sock, bool isOOB)
 {
-    fd_set readFds, writeFds, exceptFds;
-    struct timeval toWait = { 0, 0 };
-    // Mac OS X requires the usec field to be less than a million
-    toWait.tv_sec = maxMillisecs / 1000;
-    toWait.tv_usec = (maxMillisecs % 1000) * 1000;
-    FD_ZERO(&readFds);
-    FD_ZERO(&writeFds);
-    FD_ZERO(&exceptFds);
-    FD_SET(m_sock, m_isOOB ? &exceptFds : &readFds);
-    int result = select(FD_SETSIZE, &readFds, &writeFds, &exceptFds, &toWait);
-    ASSERT(result >= 0 || errno == EINTR); // The only "error" should be an interrupt.
+    if (isOOB) SetExcept(sock); else SetRead(sock);
 }
 
+// Wait for a socket to be free to write.
+class WaitNetSend: public WaitSelect {
+public:
+    WaitNetSend(SOCKET sock) { SetWrite(sock); }
+};
+
 Handle Net_dispatch_c(TaskData *taskData, Handle args, Handle code)
 {
-    int c = get_C_int(taskData, DEREFWORDHANDLE(code));
-TryAgain:
+    unsigned c = get_C_unsigned(taskData, DEREFWORDHANDLE(code));
+TryAgain: // Used for various retries.
+          // N.B.  If we call ThreadPause etc we may GC.  We MUST reload any handles so for
+          // safety we always come back here.
     switch (c)
     {
     case 0:
@@ -680,7 +703,7 @@ TryAgain:
             if (getpeername(strm->device.sock, &sockA, &size) != 0)
                 raise_syscall(taskData, "getpeername failed", GETERROR);
             /* Addresses are treated as strings. */
-            return(SAVE(Buffer_to_Poly(taskData, (char*)&sockA, size)));
+            return(SAVE(C_string_to_Poly(taskData, (char*)&sockA, size)));
         }
 
     case 38: /* Get socket name. */
@@ -691,7 +714,7 @@ TryAgain:
             if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
             if (getsockname(strm->device.sock, &sockA, &size) != 0)
                 raise_syscall(taskData, "getsockname failed", GETERROR);
-            return(SAVE(Buffer_to_Poly(taskData, (char*)&sockA, size)));
+            return(SAVE(C_string_to_Poly(taskData, (char*)&sockA, size)));
         }
 
     case 39: /* Return the address family from an address. */
@@ -710,7 +733,7 @@ TryAgain:
             sockaddr.sin_port = htons(get_C_ushort(taskData, DEREFHANDLE(args)->Get(0)));
             sockaddr.sin_addr.s_addr =
                 htonl(get_C_unsigned(taskData, DEREFHANDLE(args)->Get(1)));
-            return(SAVE(Buffer_to_Poly(taskData, (char*)&sockaddr, sizeof(sockaddr))));
+            return(SAVE(C_string_to_Poly(taskData, (char*)&sockaddr, sizeof(sockaddr))));
         }
 
     case 41: /* Return port number from an internet socket address.
@@ -807,7 +830,8 @@ TryAgain:
                            suspend this process and try again later. */
                         if (c == 46 /* blocking version. */) {
                             WaitNet waiter(strm->device.sock);
-                            processes->BlockAndRestart(taskData, &waiter, false, POLY_SYS_network);
+                            processes->ThreadPauseForIO(taskData, &waiter);
+                            goto TryAgain;
                         }
                         /* else drop through. */
                     default:
@@ -815,7 +839,7 @@ TryAgain:
                     }
                 }
 
-                addrHandle = SAVE(Buffer_to_Poly(taskData, (char*)&resultAddr, addrLen));
+                addrHandle = SAVE(C_string_to_Poly(taskData, (char*)&resultAddr, addrLen));
                 newStrm = &basic_io_vector[stream_no];
                 newStrm->device.sock = result;
                 newStrm->ioBits =
@@ -847,70 +871,53 @@ TryAgain:
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
             PolyStringObject * psAddr = (PolyStringObject *)args->WordP()->Get(1).AsObjPtr();
             struct sockaddr *psock = (struct sockaddr *)&psAddr->chars;
-            int res;
             if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
             /* In Windows, and possibly also in Unix, if we have
                received a previous EWOULDBLOCK we have to use "select"
                to tell us whether the connection actually succeeded. */
             while (1)
             {
-                if (strm->ioBits & IO_BIT_INPROGRESS)
+                int res = connect(strm->device.sock, psock, (int)psAddr->length);
+                if (res == 0) return Make_arbitrary_precision(taskData, 0); /* OK */
+                /* It isn't clear that EINTR can ever occur with
+                    connect, but just to be safe, we retry. */
+                int err = GETERROR;
+                if ((err == EWOULDBLOCK || err == EINPROGRESS) && c == 48 /*blocking version*/)
+                    break; // It's in progress and we need to wait for completion
+                else if (err != EINTR)
+                    raise_syscall(taskData, "connect failed", err);
+                /* else try again. */
+            }
+
+
+            while (1)
+            {
+                // ThreadPause may GC.  We need to reload the socket for security.
+                strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
+                if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
+                SOCKET sock = strm->device.sock;
+                /* In Windows failure is indicated by the bit being set in
+                    the exception set rather than the write set. */
+                WaitSelect waiter;
+                waiter.SetWrite(sock);
+                waiter.SetExcept(sock);
+                processes->ThreadPauseForIO(taskData, &waiter);
+
+                if (waiter.SelectResult() < 0)
                 {
-                    fd_set read_fds, write_fds, except_fds;
-                    struct timeval delay;
-                    int sel;
-                    SOCKET sock = strm->device.sock;
-                    FD_ZERO(&read_fds);
-                    FD_ZERO(&write_fds);
-                    FD_ZERO(&except_fds);
-                    FD_SET(sock, &write_fds);
-                    FD_SET(sock, &except_fds);
-                    delay.tv_sec  = 0; /* Poll. */
-                    delay.tv_usec = 0;
-                    /* In Windows failure is indicated by the bit being set in
-                       the exception set rather than the write set. */
-                    sel = select(FD_SETSIZE,&read_fds,&write_fds,&except_fds,&delay);
-                    if (sel < 0)
-                    {
-                        int err = GETERROR;
-                        if (err != EINTR)
-                            raise_syscall(taskData, "select failed", err);
-                        /* else continue */
-                    }
-                    else if (sel == 0)
-                    {
-                        /* Nothing yet */
-                        processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_network);
-                            /* -1 => not for reading. */
-                    }
-                    else /* Definite result. */
-                    {
-                        int result = 0;
-                        socklen_t len = sizeof(result);
-                        strm->ioBits &= ~IO_BIT_INPROGRESS; /* No longer in progress. */
-                        if (getsockopt(sock, SOL_SOCKET, SO_ERROR, (char*)&result, &len) != 0
-                            || result != 0)
-                            raise_syscall(taskData, "connect failed", MAPERROR(result));
-                        return Make_arbitrary_precision(taskData, 0); /* Success. */
-                    }
+                    int err = waiter.SelectError();
+                    if (err != EINTR)
+                        raise_syscall(taskData, "select failed", err);
+                    /* else continue */
                 }
-                else
+                else if (waiter.SelectResult() != 0) /* Definite result. */
                 {
-                    int err;
-                    res = connect(strm->device.sock, psock, (int)psAddr->length);
-                    if (res == 0) return Make_arbitrary_precision(taskData, 0); /* OK */
-                    /* It isn't clear that EINTR can ever occur with
-                       connect, but just to be safe, we retry. */
-                    err = GETERROR;
-                    if ((err == EWOULDBLOCK || err == EINPROGRESS) && c == 48 /*blocking version*/)
-                    {
-                        strm->ioBits |= IO_BIT_INPROGRESS;
-                        processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_network);
-                            /* -1 => not for reading. */
-                    }
-                    else if (err != EINTR)
-                        raise_syscall(taskData, "connect failed", err);
-                    /* else try again. */
+                    int result = 0;
+                    socklen_t len = sizeof(result);
+                    if (getsockopt(sock, SOL_SOCKET, SO_ERROR, (char*)&result, &len) != 0
+                        || result != 0)
+                        raise_syscall(taskData, "connect failed", MAPERROR(result));
+                    return Make_arbitrary_precision(taskData, 0); /* Success. */
                 }
             }
         }
@@ -949,11 +956,15 @@ TryAgain:
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
             PolyWord pBase = DEREFHANDLE(args)->Get(1);
             char    ch, *base;
-            unsigned int offset = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(2));
-            unsigned int length = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(3));
+            POLYUNSIGNED offset = getPolyUnsigned(taskData, DEREFHANDLE(args)->Get(2));
+#if(defined(_WIN32) && ! defined(_CYGWIN))
+            int length = get_C_int(taskData, DEREFHANDLE(args)->Get(3));
+#else
+            ssize_t length = getPolyUnsigned(taskData, DEREFHANDLE(args)->Get(3));
+#endif
             unsigned int dontRoute = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(4));
             unsigned int outOfBand = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(5));
-            int flags = 0, sent;
+            int flags = 0;
             if (dontRoute != 0) flags |= MSG_DONTROUTE;
             if (outOfBand != 0) flags |= MSG_OOB;
             if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
@@ -970,6 +981,11 @@ TryAgain:
             while (1)
             {
                 int err;
+#if(defined(_WIN32) && ! defined(_CYGWIN))
+                int sent;
+#else
+                ssize_t sent;
+#endif
                 sent = send(strm->device.sock, base+offset, length, flags);
                 /* It isn't clear that EINTR can ever occur with
                    send but just to be safe we deal with that case and
@@ -979,9 +995,10 @@ TryAgain:
                 err = GETERROR;
                 if (err == EWOULDBLOCK && c == 51 /* blocking */)
                 {
-                    processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_network);
-                        /* -1 => not for reading */
-                    ASSERT(0); /* Must not have returned. */
+                    WaitNetSend waiter(strm->device.sock);
+                    processes->ThreadPauseForIO(taskData, &waiter);
+                    // It is NOT safe to just loop here.  We may have GCed.
+                    goto TryAgain;
                 }
                 else if (err != EINTR)
                     raise_syscall(taskData, "send failed", err);
@@ -998,11 +1015,15 @@ TryAgain:
             PolyStringObject * psAddr = (PolyStringObject *)args->WordP()->Get(1).AsObjPtr();
             PolyWord pBase = DEREFHANDLE(args)->Get(2);
             char    ch, *base;
-            unsigned int offset = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(3));
-            unsigned int length = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(4));
+            POLYUNSIGNED offset = getPolyUnsigned(taskData, DEREFHANDLE(args)->Get(3));
+#if(defined(_WIN32) && ! defined(_CYGWIN))
+            int length = get_C_int(taskData, DEREFHANDLE(args)->Get(4));
+#else
+            size_t length = getPolyUnsigned(taskData, DEREFHANDLE(args)->Get(4));
+#endif
             unsigned int dontRoute = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(5));
             unsigned int outOfBand = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(6));
-            int flags = 0, sent;
+            int flags = 0;
             if (dontRoute != 0) flags |= MSG_DONTROUTE;
             if (outOfBand != 0) flags |= MSG_OOB;
             if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
@@ -1019,6 +1040,11 @@ TryAgain:
             while (1)
             {
                 int err;
+#if(defined(_WIN32) && ! defined(_CYGWIN))
+                int sent;
+#else
+                ssize_t sent;
+#endif
                 sent = sendto(strm->device.sock, base+offset, length, flags,
                             (struct sockaddr *)psAddr->chars, (int)psAddr->length);
                 /* It isn't clear that EINTR can ever occur with
@@ -1029,8 +1055,10 @@ TryAgain:
                 err = GETERROR;
                 if (err == EWOULDBLOCK && c == 52 /* blocking */)
                 {
-                    processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_network);
-                    ASSERT(0); /* Must not have returned. */
+                    WaitNetSend waiter(strm->device.sock);
+                    processes->ThreadPauseForIO(taskData, &waiter);
+                    // It is NOT safe to just loop here.  We may have GCed.
+                    goto TryAgain;
                 }
                 else if (err != EINTR)
                     raise_syscall(taskData, "sendto failed", err);
@@ -1045,17 +1073,26 @@ TryAgain:
         {
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
             char *base = (char*)DEREFHANDLE(args)->Get(1).AsObjPtr()->AsBytePtr();
-            unsigned int offset = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(2));
-            unsigned int length = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(3));
+            POLYUNSIGNED offset = getPolyUnsigned(taskData, DEREFHANDLE(args)->Get(2));
+#if(defined(_WIN32) && ! defined(_CYGWIN))
+            int length = get_C_int(taskData, DEREFHANDLE(args)->Get(3));
+#else
+            size_t length = getPolyUnsigned(taskData, DEREFHANDLE(args)->Get(3));
+#endif
             unsigned int peek = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(4));
             unsigned int outOfBand = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(5));
-            int flags = 0, recvd;
+            int flags = 0;
             if (peek != 0) flags |= MSG_PEEK;
             if (outOfBand != 0) flags |= MSG_OOB;
             if (strm == NULL) raise_syscall(taskData, "Stream is closed", EBADF);
 
             while (1) {
                 int err;
+#if(defined(_WIN32) && ! defined(_CYGWIN))
+                int recvd;
+#else
+                ssize_t recvd;
+#endif
                 recvd = recv(strm->device.sock, base+offset, length, flags);
                 err = GETERROR;
                 if (recvd != SOCKET_ERROR) { /* OK. */
@@ -1068,8 +1105,9 @@ TryAgain:
                 {
                     /* Block until something arrives. */
                     WaitNet waiter(strm->device.sock, outOfBand != 0);
-                    processes->BlockAndRestart(taskData, &waiter, false, POLY_SYS_network);
-                    ASSERT(0); /* Must not have returned. */
+                    processes->ThreadPauseForIO(taskData, &waiter);
+                    // It is NOT safe to just loop here.  We may have GCed.
+                    goto TryAgain;
                 }
                 else if (err != EINTR)
                     raise_syscall(taskData, "recv failed", err);
@@ -1086,11 +1124,15 @@ TryAgain:
         {
             PIOSTRUCT strm = get_stream(DEREFHANDLE(args)->Get(0).AsObjPtr());
             char *base = (char*)DEREFHANDLE(args)->Get(1).AsObjPtr()->AsBytePtr();
-            unsigned int offset = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(2));
-            unsigned int length = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(3));
+            POLYUNSIGNED offset = getPolyUnsigned(taskData, DEREFHANDLE(args)->Get(2));
+#if(defined(_WIN32) && ! defined(_CYGWIN))
+            int length = get_C_int(taskData, DEREFHANDLE(args)->Get(3));
+#else
+            size_t length = getPolyUnsigned(taskData, DEREFHANDLE(args)->Get(3));
+#endif
             unsigned int peek = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(4));
             unsigned int outOfBand = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(5));
-            int flags = 0, recvd;
+            int flags = 0;
             socklen_t addrLen;
             struct sockaddr resultAddr;
 
@@ -1100,6 +1142,11 @@ TryAgain:
 
             while (1) {
                 int err;
+#if(defined(_WIN32) && ! defined(_CYGWIN))
+                int recvd;
+#else
+                ssize_t recvd;
+#endif
                 recvd = recvfrom(strm->device.sock, base+offset,
                                  length, flags, &resultAddr, &addrLen);
                 err = GETERROR;
@@ -1108,7 +1155,7 @@ TryAgain:
                     Handle addrHandle, lengthHandle, pair;
                     if (recvd > (int)length) recvd = length;
                     lengthHandle = Make_arbitrary_precision(taskData, recvd);
-                    addrHandle = SAVE(Buffer_to_Poly(taskData, (char*)&resultAddr, addrLen));
+                    addrHandle = SAVE(C_string_to_Poly(taskData, (char*)&resultAddr, addrLen));
                     pair = ALLOC(2);
                     DEREFHANDLE(pair)->Set(0, DEREFWORDHANDLE(lengthHandle));
                     DEREFHANDLE(pair)->Set(1, DEREFWORDHANDLE(addrHandle));
@@ -1117,8 +1164,9 @@ TryAgain:
                 if (err == EWOULDBLOCK && c == 54 /* blocking */)
                 {
                     WaitNet waiter(strm->device.sock, outOfBand != 0);
-                    processes->BlockAndRestart(taskData, &waiter, false, POLY_SYS_network);
-                    ASSERT(0); /* Must not have returned. */
+                    processes->ThreadPauseForIO(taskData, &waiter);
+                    // It is NOT safe to just loop here.  We may have GCed.
+                    goto TryAgain;
                 }
                 else if (err != EINTR)
                     raise_syscall(taskData, "recvfrom failed", err);
@@ -1202,7 +1250,7 @@ TryAgain:
             POLYUNSIGNED length = Poly_string_to_C(DEREFWORD(args), addr.sun_path, sizeof(addr.sun_path));
             if (length > (int)sizeof(addr.sun_path))
                 raise_syscall(taskData, "Address too long", ENAMETOOLONG);
-            return SAVE(Buffer_to_Poly(taskData, (char*)&addr, sizeof(addr)));
+            return SAVE(C_string_to_Poly(taskData, (char*)&addr, sizeof(addr)));
         }
 #endif
 
@@ -1238,31 +1286,6 @@ TryAgain:
     }
 }
 
-/* "Polymorphic" function to generate a list. */
-static Handle makeList(TaskData *taskData, int count, char *p, int size, void *arg,
-                       Handle (mkEntry)(TaskData *, void*, char*))
-{
-    Handle saved = taskData->saveVec.mark();
-    Handle list = SAVE(ListNull);
-    /* Start from the end of the list. */
-    p += count*size;
-    while (count > 0)
-    {
-        Handle value, next;
-        p -= size; /* Back up to the last entry. */
-        value = mkEntry(taskData, arg, p);
-        next  = ALLOC(SIZEOF(ML_Cons_Cell));
-
-        DEREFLISTHANDLE(next)->h = DEREFWORDHANDLE(value); 
-        DEREFLISTHANDLE(next)->t = DEREFLISTHANDLE(list);
-
-        taskData->saveVec.reset(saved);
-        list = SAVE(DEREFHANDLE(next));
-        count--;
-    }
-    return list;
-}
-
 static Handle mkAddr(TaskData *taskData, void *arg, char *p)
 {
     int j;
@@ -1460,6 +1483,7 @@ static Handle getSelectResult(TaskData *taskData, Handle args, int offset, fd_se
    minus one.  */
 static Handle selectCall(TaskData *taskData, Handle args, int blockType)
 {
+    TryAgain:
     // We should check for interrupts even if we're not going to block.
     processes->TestAnyEvents(taskData);
     fd_set readers, writers, excepts;
@@ -1504,11 +1528,12 @@ static Handle selectCall(TaskData *taskData, Handle args, int blockType)
             {
             /* The time argument is an absolute time. */
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
+            Handle hSave = taskData->saveVec.mark();
             FILETIME ftTime, ftNow;
             /* Get the file time. */
-            get_C_pair(taskData, DEREFHANDLE(args)->Get(3),
-                &ftTime.dwHighDateTime, &ftTime.dwLowDateTime);
+            getFileTimeFromArb(taskData, taskData->saveVec.push(DEREFHANDLE(args)->Get(3)), &ftTime);
             GetSystemTimeAsFileTime(&ftNow);
+            taskData->saveVec.reset(hSave);
             /* If the timeout time is earlier than the current time
                we must return, otherwise we block. */
             if (CompareFileTime(&ftTime, &ftNow) <= 0)
@@ -1516,6 +1541,7 @@ static Handle selectCall(TaskData *taskData, Handle args, int blockType)
             /* else drop through and block. */
 #else /* Unix */
             struct timeval tv;
+            Handle hSave = taskData->saveVec.mark();
             /* We have a value in microseconds.  We need to split
                it into seconds and microseconds. */
             Handle hTime = SAVE(DEREFWORDHANDLE(args)->Get(3));
@@ -1524,6 +1550,7 @@ static Handle selectCall(TaskData *taskData, Handle args, int blockType)
                 get_C_ulong(taskData, DEREFWORDHANDLE(div_longc(taskData, hMillion, hTime)));
             unsigned long usecs =
                 get_C_ulong(taskData, DEREFWORDHANDLE(rem_longc(taskData, hMillion, hTime)));
+            taskData->saveVec.reset(hSave);
             /* If the timeout time is earlier than the current time
                we must return, otherwise we block. */
             if (gettimeofday(&tv, NULL) != 0)
@@ -1535,8 +1562,8 @@ static Handle selectCall(TaskData *taskData, Handle args, int blockType)
 #endif
         }
         case 1: /* Block until one of the descriptors is ready. */
-            processes->BlockAndRestart(taskData, NULL, false, POLY_SYS_network);
-            /*NOTREACHED*/
+            processes->ThreadPause(taskData);
+            goto TryAgain;
         case 2: /* Just a simple poll - drop through. */
             break;
         }
@@ -1567,13 +1594,8 @@ static Networking networkingModule;
 void Networking::Init(void)
 {
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
-#ifdef USEWINSOCK2
 #define WINSOCK_MAJOR_VERSION   2
 #define WINSOCK_MINOR_VERSION   2
-#else
-#define WINSOCK_MAJOR_VERSION   1
-#define WINSOCK_MINOR_VERSION   1
-#endif
     WSADATA wsaData;
     WORD wVersion = MAKEWORD(WINSOCK_MINOR_VERSION, WINSOCK_MAJOR_VERSION);
     /* Initialise the system and check that the version it supplied
