@@ -2,12 +2,11 @@
     Title:  No return header.
     Author: David C.J. Matthews
 
-    Copyright (c) 2006 David C.J. Matthews
+    Copyright (c) 2006, 2015 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -25,15 +24,12 @@
 
 /* The exception functions don't return but instead raise exceptions.  This macro
    tells the compiler about this and prevents spurious errors.  */
-#if defined(__GNUC__) || defined(__attribute__)
-#define NORETURNFN(x)   x __attribute__((noreturn))
-#define NORET        0
-#elif (defined(_MSC_EXTENSIONS) && (_MSC_VER >= 1200))
+#if (defined(_MSC_EXTENSIONS) && (_MSC_VER >= 1200))
 #define NORETURNFN(x) __declspec(noreturn) x
-#define NORET        0
+#elif defined(__GNUC__) || defined(__attribute__)
+#define NORETURNFN(x)   x __attribute__((noreturn))
 #else
-#define NORETURNFN(x)
-#define NORET        return 0
+#define NORETURNFN(x) x
 #endif
 
 
