@@ -2,13 +2,12 @@
     Title:     Export and import memory in a portable format
     Author:    David C. J. Matthews.
 
-    Copyright (c) 2006-7 David C. J. Matthews
+    Copyright (c) 2006-7, 2015-6 David C. J. Matthews
 
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -88,7 +87,7 @@ unsigned long PExport::getIndex(PolyObject *p)
     {
         ASSERT(lower < upper);
         unsigned long middle = (lower+upper)/2;
-        ASSERT(middle >= 0 && middle < nObjects);
+        ASSERT(middle < nObjects);
         if (p < pMap[middle])
         {
             // Use lower to middle
@@ -123,7 +122,7 @@ void PExport::printAddress(void *p)
         POLYUNSIGNED byteOffset = (char*)p - (char*)memTable[area].mtAddr;
         unsigned ioEntry = (unsigned)(byteOffset / (ioSpacing*sizeof(PolyWord)));
         unsigned ioOffset = (unsigned)(byteOffset - ioEntry * (ioSpacing*sizeof(PolyWord)));
-        ASSERT(ioEntry >= 0 && ioEntry < POLY_SYS_vecsize);
+        ASSERT(ioEntry < POLY_SYS_vecsize);
         if (ioOffset == 0)
            fprintf(exportFile, "I%d", ioEntry);
         else
@@ -532,7 +531,7 @@ bool PImport::GetValue(PolyWord *result)
         /* IO entry number. */
         POLYUNSIGNED j;
         fscanf(f, "%" POLYUFMT, &j);
-        ASSERT(j >= 0 && j < POLY_SYS_vecsize);
+        ASSERT(j < POLY_SYS_vecsize);
         *result = (PolyObject*)&gMem.ioSpace->bottom[j * IO_SPACING];
     }
     else if (ch == 'J')
@@ -540,7 +539,7 @@ bool PImport::GetValue(PolyWord *result)
         /* IO entry number with offset. */
         POLYUNSIGNED j, offset;
         fscanf(f, "%" POLYUFMT "+%" POLYUFMT, &j, &offset);
-        ASSERT(j >= 0 && j < POLY_SYS_vecsize);
+        ASSERT(j < POLY_SYS_vecsize);
         PolyWord base = (PolyObject*)&gMem.ioSpace->bottom[j * IO_SPACING];
         *result = PolyWord::FromCodePtr(base.AsCodePtr() + offset);
     }
@@ -827,65 +826,15 @@ bool PImport::DoImport()
                         do ch = getc(f); while (ch == ' ');
                     }
                 }
-                break;
-            }
-
-        case 'Q': /* Stack segment. */
-            {
-                StackObject *s = (StackObject*)p;
-                POLYUNSIGNED n;
-                POLYUNSIGNED length = p->Length();
-                fscanf(f, "%" POLYUFMT, &nWords);
-                ch = getc(f); ASSERT(ch == '|');
-
-                /* Standard fields: size, pc, sp, hr. */
-                fscanf(f, "%" POLYUFMT, &s->p_space);
-                ch = getc(f); ASSERT(ch == ',');
-                if (! ReadValue(p, (PolyWord*)&s->p_pc - (PolyWord*)p))
-                    return false;
-                ch = getc(f); ASSERT(ch == ',');
-                if (! ReadValue(p, (PolyWord*)&s->p_sp - (PolyWord*)p))
-                    return false;
-                ch = getc(f); ASSERT(ch == ',');
-                if (! ReadValue(p, (PolyWord*)&s->p_hr - (PolyWord*)p))
-                    return false;
-
-                /* Checked registers. */
-                fscanf(f, "%" POLYUFMT, &n);
-                s->p_nreg = n;
-                ch = getc(f); ASSERT(ch == '|');
-                for (i = 0; i < n; i++)
+                // Adjust the byte count.  This is only necessary when importing the interpreted
+                // code into a machine with a different endian-ness from the exporter.
                 {
-                    if (! ReadValue(p, &s->p_reg[i] - (PolyWord*)p))
-                        return false;
-                    ch = getc(f);
-                    ASSERT((ch == ',' && i < n-1) ||
-                           (ch == ' ' && i == n-1));
+                    POLYUNSIGNED l = 0;
+                    while (l < length && p->Get(l) != PolyWord::FromUnsigned(0)) l++;
+                    ASSERT(l < length);
+                    l++;
+                    p->Set(l, PolyWord::FromUnsigned(l*sizeof(PolyWord)));
                 }
-                /* Unchecked registers. */
-                fscanf(f, "%" POLYUFMT, &n);
-                s->p_reg[i] = PolyWord::FromUnsigned(n);
-                ch = getc(f); ASSERT(ch == '|');
-                for (i = 0; i < n; i++)
-                {
-                    POLYSIGNED n;
-                    fscanf(f, "%" POLYSFMT, &n);
-                    s->p_reg[s->p_nreg+i+1] = PolyWord::FromSigned(n);
-                    ch = getc(f);
-                }
-                /* Stack values. */
-                fscanf(f, "%" POLYUFMT, &n);
-                ASSERT(n == length - (s->p_sp-(PolyWord*)p));
-                ch = getc(f); ASSERT(ch == '|');
-                for (i = 0; i < n; i++)
-                {
-                    if(! ReadValue(p, length-n+i))
-                        return false;
-                    ch = getc(f);
-                    ASSERT((ch == ',' && i < n-1) ||
-                           (ch == '\n' && i == n-1));
-                }
-
                 break;
             }
 
@@ -898,15 +847,24 @@ bool PImport::DoImport()
 }
 
 // Import a file in the portable format and return a pointer to the root object.
-PolyObject *ImportPortable(const char *fileName)
+PolyObject *ImportPortable(const TCHAR *fileName)
 {
     PImport pImport;
+#if (defined(_WIN32) && defined(UNICODE))
+    pImport.f = _wfopen(fileName, L"r");
+    if (pImport.f == 0)
+    {
+        fprintf(stderr, "Unable to open file: %S\n", fileName);
+        return 0;
+    }
+#else
     pImport.f = fopen(fileName, "r");
     if (pImport.f == 0)
     {
         fprintf(stderr, "Unable to open file: %s\n", fileName);
         return 0;
     }
+#endif
     if (pImport.DoImport())
         return pImport.Root();
     else
