@@ -2,12 +2,11 @@
     Title:      Thread functions
     Author:     David C.J. Matthews
 
-    Copyright (c) 2007,2008,2013 David C.J. Matthews
+    Copyright (c) 2007,2008,2013-15 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -93,6 +92,10 @@
 #include <sys/sysctl.h>
 #endif
 
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#include <tchar.h>
+#endif
+
 #include <new>
 
 /************************************************************************
@@ -141,50 +144,6 @@
 #define PFLAG_INTMASK       6   // Mask of the above bits
 
 
-// Other threads may make requests to a thread.
-typedef enum {
-    kRequestNone = 0, // Increasing severity
-    kRequestInterrupt = 1,
-    kRequestKill = 2
-} ThreadRequests;
-
-class ProcessTaskData: public TaskData
-{
-public:
-    ProcessTaskData();
-    ~ProcessTaskData();
-
-    virtual void Lock(void) {}
-    virtual void Unlock(void) {}
-
-    virtual void GarbageCollect(ScanAddress *process);
-
-    // If a thread has to block it will block on this.
-    PCondVar threadLock;
-    // External requests made are stored here until they
-    // can be actioned.
-    ThreadRequests requests;
-    // Pointer to the mutex when blocked. Set to NULL when it doesn't apply.
-    PolyObject *blockMutex;
-    // This is set to false when a thread blocks or enters foreign code,
-    // While it is true the thread can manipulate ML memory so no other
-    // thread can garbage collect.
-    bool inMLHeap;
-
-    // In Linux, at least, we need to run a separate timer in each thread
-    bool runningProfileTimer;
-
-#ifdef HAVE_WINDOWS_H
-    LONGLONG lastCPUTime; // Used for profiling
-#endif
-#ifdef HAVE_PTHREAD
-    bool threadExited;
-#endif
-#ifdef HAVE_WINDOWS_H
-    HANDLE threadHandle;
-#endif
-};
-
 class Processes: public ProcessExternal, public RtsModule
 {
 public:
@@ -199,7 +158,6 @@ public:
     // Called when a thread has completed - doesn't return.
     virtual NORETURNFN(void ThreadExit(TaskData *taskData));
 
-    void BlockAndRestart(TaskData *taskData, Waiter *pWait, bool posixInterruptable, int ioCall);
     // Called when a thread may block.  Returns some time later when perhaps
     // the input is available.
     virtual void ThreadPauseForIO(TaskData *taskData, Waiter *pWait);
@@ -212,8 +170,8 @@ public:
     virtual bool ForkFromRTS(TaskData *taskData, Handle proc, Handle arg);
     // Create a new thread.  The "args" argument is only used for threads
     // created in the RTS by the signal handler.
-    Handle ForkThread(ProcessTaskData *taskData, Handle threadFunction,
-                    Handle args, PolyWord flags);
+    Handle ForkThread(TaskData *taskData, Handle threadFunction,
+                    Handle args, PolyWord flags, PolyWord stacksize);
     // Process general RTS requests from ML.
     Handle ThreadDispatch(TaskData *taskData, Handle args, Handle code);
 
@@ -237,7 +195,7 @@ public:
 
     // Set a thread to be interrupted or killed.  Wakes up the
     // thread if necessary.  MUST be called with taskArrayLock held.
-    void MakeRequest(ProcessTaskData *p, ThreadRequests request);
+    void MakeRequest(TaskData *p, ThreadRequests request);
 
     // Profiling control.
     virtual void StartProfiling(void);
@@ -257,7 +215,7 @@ public:
 
     // Find a task that matches the specified identifier and returns
     // it if it exists.  MUST be called with taskArrayLock held.
-    ProcessTaskData *TaskForIdentifier(Handle taskId);
+    TaskData *TaskForIdentifier(Handle taskId);
 
     // Signal handling support.  The ML signal handler thread blocks until it is
     // woken up by the signal detection thread.
@@ -272,7 +230,7 @@ public:
     bool singleThreaded;
 
     // Each thread has an entry in this array.
-    ProcessTaskData **taskArray;
+    TaskData **taskArray;
     unsigned taskArraySize; // Current size of the array.
 
     /* schedLock: This lock must be held when making scheduling decisions.
@@ -318,7 +276,7 @@ public:
     LONGLONG lastCPUTime; // CPU used by main thread.
 #endif
 
-    ProcessTaskData *sigTask;  // Pointer to current signal task.
+    TaskData *sigTask;  // Pointer to current signal task.
 };
 
 // Global process data.
@@ -349,21 +307,21 @@ static POLYUNSIGNED ThreadAttrs(TaskData *taskData)
 
 // Called from interface vector.  Generally the assembly code will be
 // used instead of this.
-Handle AtomicIncrement(TaskData *taskData, Handle mutexp)
+Handle ProcessAtomicIncrement(TaskData *taskData, Handle mutexp)
 {
-    return machineDependent->AtomicIncrement(taskData, mutexp);
+    return taskData->AtomicIncrement(mutexp);
 }
 
 // Called from interface vector.  Generally the assembly code will be
 // used instead of this.
-Handle AtomicDecrement(TaskData *taskData, Handle mutexp)
+Handle ProcessAtomicDecrement(TaskData *taskData, Handle mutexp)
 {
-    return machineDependent->AtomicDecrement(taskData, mutexp);
+    return taskData->AtomicDecrement(mutexp);
 }
 
-Handle AtomicReset(TaskData *taskData, Handle mutexp)
+Handle ProcessAtomicReset(TaskData *taskData, Handle mutexp)
 {
-    machineDependent->AtomicReset(taskData, mutexp);
+    taskData->AtomicReset(mutexp);
     return SAVE(TAGGED(0)); // Push the unit result
 }
 
@@ -383,8 +341,8 @@ Handle ThreadDispatch(TaskData *taskData, Handle args, Handle code)
 
 Handle Processes::ThreadDispatch(TaskData *taskData, Handle args, Handle code)
 {
-    int c = get_C_int(taskData, DEREFWORDHANDLE(code));
-    ProcessTaskData *ptaskData = (ProcessTaskData *)taskData;
+    unsigned c = get_C_unsigned(taskData, DEREFWORDHANDLE(code));
+    TaskData *ptaskData = taskData;
     switch (c)
     {
     case 1: /* A mutex was locked i.e. the count was ~1 or less.  We will have set it to
@@ -450,7 +408,7 @@ Handle Processes::ThreadDispatch(TaskData *taskData, Handle args, Handle code)
             // Unlock any waiters.
             for (unsigned i = 0; i < taskArraySize; i++)
             {
-                ProcessTaskData *p = taskArray[i];
+                TaskData *p = taskArray[i];
                 // If the thread is blocked on this mutex we can signal the thread.
                 if (p && p->blockMutex == DEREFHANDLE(args))
                     p->threadLock.Signal();
@@ -480,10 +438,7 @@ Handle Processes::ThreadDispatch(TaskData *taskData, Handle args, Handle code)
             // On Windows it is the number of 100ns units since the epoch
             FILETIME tWake;
             if (! isInfinite)
-            {
-                get_C_pair(taskData, DEREFWORDHANDLE(wakeTime),
-                    (unsigned long*)&tWake.dwHighDateTime, (unsigned long*)&tWake.dwLowDateTime);
-            }
+                getFileTimeFromArb(taskData, wakeTime, &tWake);
 #else
             // Unix style times.
             struct timespec tWake;
@@ -500,15 +455,15 @@ Handle Processes::ThreadDispatch(TaskData *taskData, Handle args, Handle code)
             schedLock.Lock();
             // Atomically release the mutex.  This is atomic because we hold schedLock
             // so no other thread can call signal or broadcast.
-            Handle decrResult = machineDependent->AtomicIncrement(taskData, mutexH);
+            Handle decrResult = taskData->AtomicIncrement(mutexH);
             if (UNTAGGED(decrResult->Word()) != 1)
             {
-                machineDependent->AtomicReset(taskData, mutexH);
+                taskData->AtomicReset(mutexH);
                 // The mutex was locked so we have to release any waiters.
                 // Unlock any waiters.
                 for (unsigned i = 0; i < taskArraySize; i++)
                 {
-                    ProcessTaskData *p = taskArray[i];
+                    TaskData *p = taskArray[i];
                     // If the thread is blocked on this mutex we can signal the thread.
                     if (p && p->blockMutex == DEREFHANDLE(mutexH))
                         p->threadLock.Signal();
@@ -544,7 +499,7 @@ Handle Processes::ThreadDispatch(TaskData *taskData, Handle args, Handle code)
             // Acquire the schedLock first.  This ensures that this is
             // atomic with respect to waiting.
             schedLock.Lock();
-            ProcessTaskData *p = TaskForIdentifier(args);
+            TaskData *p = TaskForIdentifier(args);
             if (p && p->threadObject == args->WordP())
             {
                 POLYUNSIGNED attrs = ThreadAttrs(p) & PFLAG_INTMASK;
@@ -563,12 +518,14 @@ Handle Processes::ThreadDispatch(TaskData *taskData, Handle args, Handle code)
 
     case 7: // Fork a new thread.  The arguments are the function to run and the attributes.
             return ForkThread(ptaskData, SAVE(args->WordP()->Get(0)),
-                        (Handle)0, args->WordP()->Get(1));
+                        (Handle)0, args->WordP()->Get(1),
+                        // For backwards compatibility we check the length here
+                        args->WordP()->Length() <= 2 ? TAGGED(0) : args->WordP()->Get(2));
 
     case 8: // Test if a thread is active
         {
             schedLock.Lock();
-            ProcessTaskData *p = TaskForIdentifier(args);
+            TaskData *p = TaskForIdentifier(args);
             schedLock.Unlock();
             return SAVE(TAGGED(p != 0));
         }
@@ -576,7 +533,7 @@ Handle Processes::ThreadDispatch(TaskData *taskData, Handle args, Handle code)
     case 9: // Send an interrupt to a specific thread
         {
             schedLock.Lock();
-            ProcessTaskData *p = TaskForIdentifier(args);
+            TaskData *p = TaskForIdentifier(args);
             if (p) MakeRequest(p, kRequestInterrupt);
             schedLock.Unlock();
             if (p == 0)
@@ -594,13 +551,13 @@ Handle Processes::ThreadDispatch(TaskData *taskData, Handle args, Handle code)
         // These will be handled "soon" but if we have just switched from deferring
         // interrupts this guarantees that any deferred interrupts will be handled now.
         if (ProcessAsynchRequests(taskData))
-            throw IOException(EXC_EXCEPTION);
+            throw IOException();
         return SAVE(TAGGED(0));
 
     case 12: // Kill a specific thread
         {
             schedLock.Lock();
-            ProcessTaskData *p = TaskForIdentifier(args);
+            TaskData *p = TaskForIdentifier(args);
             if (p) MakeRequest(p, kRequestKill);
             schedLock.Unlock();
             if (p == 0)
@@ -616,6 +573,20 @@ Handle Processes::ThreadDispatch(TaskData *taskData, Handle args, Handle code)
         // Returns 0 if there is any problem.
         return Make_arbitrary_precision(taskData, NumberOfPhysicalProcessors());
 
+    case 15: // Set the maximum stack size.
+        {
+            PolyWord newSize = args->Word();
+            taskData->threadObject->mlStackSize = newSize;
+            if (newSize != TAGGED(0))
+            {
+                POLYUNSIGNED current = taskData->currentStackSpace(); // Current size in words
+                POLYUNSIGNED newWords = getPolyUnsigned(taskData, newSize);
+                if (current > newWords)
+                    raise_exception0(taskData, EXC_interrupt);
+            }
+            return SAVE(TAGGED(0));
+        }
+
     default:
         {
             char msg[100];
@@ -626,20 +597,6 @@ Handle Processes::ThreadDispatch(TaskData *taskData, Handle args, Handle code)
     }
 }
 
-TaskData::TaskData(): mdTaskData(0), allocPointer(0), allocLimit(0), allocSize(MIN_HEAP_SIZE), allocCount(0),
-        stack(0), threadObject(0), signalStack(0), pendingInterrupt(false), foreignStack(TAGGED(0)),
-        inML(false)
-{
-}
-
-TaskData::~TaskData()
-{
-    if (signalStack) free(signalStack);
-    if (stack) gMem.DeleteStackSpace(stack);
-    delete(mdTaskData);
-}
-
-
 // Fill unused allocation space with a dummy object to preserve the invariant
 // that memory is always valid.
 void TaskData::FillUnusedSpace(void)
@@ -648,7 +605,10 @@ void TaskData::FillUnusedSpace(void)
         gMem.FillUnusedSpace(allocLimit, allocPointer-allocLimit); 
 }
 
-ProcessTaskData::ProcessTaskData(): requests(kRequestNone), blockMutex(0), inMLHeap(false),
+
+TaskData::TaskData(): allocPointer(0), allocLimit(0), allocSize(MIN_HEAP_SIZE), allocCount(0),
+        stack(0), threadObject(0), signalStack(0), pendingInterrupt(false), foreignStack(TAGGED(0)),
+        inML(false), requests(kRequestNone), blockMutex(0), inMLHeap(false),
         runningProfileTimer(false)
 {
 #ifdef HAVE_WINDOWS_H
@@ -662,8 +622,10 @@ ProcessTaskData::ProcessTaskData(): requests(kRequestNone), blockMutex(0), inMLH
 #endif
 }
 
-ProcessTaskData::~ProcessTaskData()
+TaskData::~TaskData()
 {
+    if (signalStack) free(signalStack);
+    if (stack) gMem.DeleteStackSpace(stack);
 #ifdef HAVE_WINDOWS_H
     if (threadHandle) CloseHandle(threadHandle);
 #endif
@@ -672,14 +634,14 @@ ProcessTaskData::~ProcessTaskData()
 
 // Find a task that matches the specified identifier and returns
 // it if it exists.  MUST be called with taskArrayLock held.
-ProcessTaskData *Processes::TaskForIdentifier(Handle taskId)
+TaskData *Processes::TaskForIdentifier(Handle taskId)
 {
     // The index is in the first word of the thread object.
     unsigned index = (unsigned)(UNTAGGED_UNSIGNED(taskId->WordP()->Get(0)));
     // Check the index is valid and matches the object stored in the table.
     if (index < taskArraySize)
     {
-        ProcessTaskData *p = taskArray[index];
+        TaskData *p = taskArray[index];
         if (p && p->threadObject == taskId->WordP())
             return p;
     }
@@ -693,7 +655,7 @@ void Processes::BroadcastInterrupt(void)
     schedLock.Lock();
     for (unsigned i = 0; i < taskArraySize; i++)
     {
-        ProcessTaskData *p = taskArray[i];
+        TaskData *p = taskArray[i];
         if (p)
         {
             POLYUNSIGNED attrs = ThreadAttrs(p);
@@ -706,13 +668,13 @@ void Processes::BroadcastInterrupt(void)
 
 // Set the asynchronous request variable for the thread.  Must be called
 // with the schedLock held.  Tries to wake the thread up if possible.
-void Processes::MakeRequest(ProcessTaskData *p, ThreadRequests request)
+void Processes::MakeRequest(TaskData *p, ThreadRequests request)
 {
     // We don't override a request to kill by an interrupt request.
     if (p->requests < request)
     {
         p->requests = request;
-        machineDependent->InterruptCode(p);
+        p->InterruptCode();
         p->threadLock.Signal();
         // Set the value in the ML object as well so the ML code can see it
         p->threadObject->requestCopy = TAGGED(request);
@@ -783,7 +745,7 @@ void Processes::ThreadReleaseMLMemory(TaskData *taskData)
 // we are woken up at the end of the GC.
 void Processes::ThreadUseMLMemoryWithSchedLock(TaskData *taskData)
 {
-    ProcessTaskData *ptaskData = (ProcessTaskData *)taskData;
+    TaskData *ptaskData = taskData;
     // If there is a request outstanding we have to wait for it to
     // complete.  We notify the root thread and wait for it.
     while (threadRequest != 0)
@@ -802,7 +764,7 @@ void Processes::ThreadUseMLMemoryWithSchedLock(TaskData *taskData)
 // that can proceed.
 void Processes::ThreadReleaseMLMemoryWithSchedLock(TaskData *taskData)
 {
-    ProcessTaskData *ptaskData = (ProcessTaskData *)taskData;
+    TaskData *ptaskData = taskData;
     ASSERT(ptaskData->inMLHeap);
     ptaskData->inMLHeap = false;
     // Put a dummy object in any unused space.  This maintains the
@@ -913,10 +875,11 @@ PolyWord *Processes::FindAllocationSpace(TaskData *taskData, POLYUNSIGNED words,
             // Try garbage-collecting.  If this failed return 0.
             if (! QuickGC(taskData, words))
             {
+                extern FILE *polyStderr;
                 if (! triedInterrupt)
                 {
                     triedInterrupt = true;
-                    fprintf(stderr,"Run out of store - interrupting threads\n");
+                    fprintf(polyStderr,"Run out of store - interrupting threads\n");
                     if (debugOptions & DEBUG_THREADS)
                         Log("THREAD: Run out of store, interrupting threads\n");
                     BroadcastInterrupt();
@@ -933,7 +896,7 @@ PolyWord *Processes::FindAllocationSpace(TaskData *taskData, POLYUNSIGNED words,
                 }
                 else {
                     // That didn't work.  Exit.
-                    fprintf(stderr,"Failed to recover - exiting\n");
+                    fprintf(polyStderr,"Failed to recover - exiting\n");
                     Exit(1); // Begins the shutdown process
                     processes->ThreadExit(taskData); // And terminate this thread.
                 }
@@ -971,28 +934,6 @@ void Processes::ThreadPauseForIO(TaskData *taskData, Waiter *pWait)
     TestAnyEvents(taskData); // Check if we've been interrupted.
 }
 
-// This is largely a legacy of the old single-thread version.  In that version there
-// was only a single C thread managing multiple ML threads (processes) so if an ML
-// thread blocked it was necessary to switch the thread and then for the C function
-// call to raise an exception to get back to ML.  
-// TODO: There's actually a race here if we have posixInterruptible set.  We
-// repeatedly come back here and if a signal happens while we're in
-// ThreadPauseForIO we will raise the exception.  If the signal happens at
-// another point we won't.
-void Processes::BlockAndRestart(TaskData *taskData, Waiter *pWait, bool posixInterruptable, int ioCall)
-{
-    if (pWait == NULL) pWait = Waiter::defaultWaiter;
-    machineDependent->SetForRetry(taskData, ioCall);
-    unsigned lastSigCount = receivedSignalCount;
-    ThreadPauseForIO(taskData, pWait);
-    // If this is an interruptible Posix function we raise an exception if
-    // there has been a signal.
-    if (posixInterruptable && lastSigCount != receivedSignalCount)
-        raise_syscall(taskData, "Call interrupted by signal", EINTR);
-    throw IOException(EXC_EXCEPTION);
-    /* NOTREACHED */
-}
-
 // Default waiter: simply wait for the time.  In the case of Windows it
 // is also woken up if the event is signalled.  In Unix it may be woken
 // up by a signal.
@@ -1090,8 +1031,7 @@ TaskData *Processes::GetTaskDataForThread(void)
 TaskData *Processes::CreateNewTaskData(Handle threadId, Handle threadFunction,
                            Handle args, PolyWord flags)
 {
-    ProcessTaskData *taskData = new ProcessTaskData;
-    taskData->mdTaskData = machineDependent->CreateTaskData();
+    TaskData *taskData = machineDependent->CreateTaskData();
 #if defined(HAVE_WINDOWS_H)
     HANDLE thisProcess = GetCurrentProcess();
     DuplicateHandle(thisProcess, GetCurrentThread(), thisProcess, 
@@ -1108,8 +1048,8 @@ TaskData *Processes::CreateNewTaskData(Handle threadId, Handle threadFunction,
 
         if (thrdIndex == taskArraySize) // Need to expand the array
         {
-            ProcessTaskData **newArray =
-                (ProcessTaskData **)realloc(taskArray, sizeof(ProcessTaskData *)*(taskArraySize+1));
+            TaskData **newArray =
+                (TaskData **)realloc(taskArray, sizeof(TaskData *)*(taskArraySize+1));
             if (newArray)
             {
                 taskArray = newArray;
@@ -1133,7 +1073,9 @@ TaskData *Processes::CreateNewTaskData(Handle threadId, Handle threadFunction,
         throw MemoryException();
     }
 
-    machineDependent->InitStackFrame(taskData, taskData->stack, threadFunction, args);
+    // TODO:  Check that there isn't a problem if we try to allocate
+    // memory here and result in a GC.
+    taskData->InitStackFrame(taskData, threadFunction, args);
 
     ThreadUseMLMemory(taskData);
 
@@ -1143,11 +1085,14 @@ TaskData *Processes::CreateNewTaskData(Handle threadId, Handle threadFunction,
         taskData->threadObject = (ThreadObject*)threadId->WordP();
     else
     {
-        taskData->threadObject = (ThreadObject*)alloc(taskData, 4, F_MUTABLE_BIT);
+        taskData->threadObject = (ThreadObject*)alloc(taskData, sizeof(ThreadObject)/sizeof(PolyWord), F_MUTABLE_BIT);
         taskData->threadObject->index = TAGGED(thrdIndex); // Set to the index
         taskData->threadObject->flags = flags != TAGGED(0) ? TAGGED(PFLAG_SYNCH): flags;
         taskData->threadObject->threadLocal = TAGGED(0); // Empty thread-local store
         taskData->threadObject->requestCopy = TAGGED(0); // Cleared interrupt state
+        taskData->threadObject->mlStackSize = TAGGED(0); // Unlimited stack size
+        for (unsigned i = 0; i < sizeof(taskData->threadObject->debuggerSlots)/sizeof(PolyWord); i++)
+            taskData->threadObject->debuggerSlots[i] = TAGGED(0);
     }
 
 #ifdef HAVE_PTHREAD
@@ -1167,7 +1112,7 @@ TaskData *Processes::CreateNewTaskData(Handle threadId, Handle threadFunction,
 #ifdef HAVE_PTHREAD
 static void *NewThreadFunction(void *parameter)
 {
-    ProcessTaskData *taskData = (ProcessTaskData *)parameter;
+    TaskData *taskData = (TaskData *)parameter;
 #ifdef HAVE_WINDOWS_H
     // Cygwin: Get the Windows thread handle in case it's needed for profiling.
     HANDLE thisProcess = GetCurrentProcess();
@@ -1180,9 +1125,9 @@ static void *NewThreadFunction(void *parameter)
     globalStats.incCount(PSC_THREADS);
     processes->ThreadUseMLMemory(taskData);
     try {
-        (void)EnterPolyCode(taskData); // Will normally (always?) call ExitThread.
+        (void)taskData->EnterPolyCode(); // Will normally (always?) call ExitThread.
     }
-    catch (KillException) {
+    catch (KillException &) {
         processesModule.ThreadExit(taskData);
     }
 
@@ -1191,15 +1136,15 @@ static void *NewThreadFunction(void *parameter)
 #elif defined(HAVE_WINDOWS_H)
 static DWORD WINAPI NewThreadFunction(void *parameter)
 {
-    ProcessTaskData *taskData = (ProcessTaskData *)parameter;
+    TaskData *taskData = (TaskData *)parameter;
     TlsSetValue(processesModule.tlsId, taskData);
     taskData->saveVec.init(); // Removal initial data
     globalStats.incCount(PSC_THREADS);
     processes->ThreadUseMLMemory(taskData);
     try {
-        (void)EnterPolyCode(taskData);
+        (void)taskData->EnterPolyCode();
     }
-    catch (KillException) {
+    catch (KillException &) {
         processesModule.ThreadExit(taskData);
     }
     return 0;
@@ -1207,15 +1152,15 @@ static DWORD WINAPI NewThreadFunction(void *parameter)
 #else
 static void NewThreadFunction(void *parameter)
 {
-    ProcessTaskData *taskData = (ProcessTaskData *)parameter;
+    TaskData *taskData = (TaskData *)parameter;
     initThreadSignals(taskData);
     taskData->saveVec.init(); // Removal initial data
     globalStats.incCount(PSC_THREADS);
     processes->ThreadUseMLMemory(taskData);
     try {
-        (void)EnterPolyCode(taskData);
+        (void)taskData->EnterPolyCode();
     }
-    catch (KillException) {
+    catch (KillException &) {
         processesModule.ThreadExit(taskData);
     }
 }
@@ -1236,22 +1181,24 @@ void Processes::BeginRootThread(PolyObject *rootFunction)
 {
     if (taskArraySize < 1)
     {
-        taskArray = (ProcessTaskData **)realloc(taskArray, sizeof(ProcessTaskData *));
+        taskArray = (TaskData **)realloc(taskArray, sizeof(TaskData *));
         if (taskArray == 0) ::Exit("Unable to create the initial thread - insufficient memory");
         taskArraySize = 1;
     }
 
     try {
         // We can't use ForkThread because we don't have a taskData object before we start
-        ProcessTaskData *taskData = new ProcessTaskData;
-        taskData->mdTaskData = machineDependent->CreateTaskData();
-        taskData->threadObject = (ThreadObject*)alloc(taskData, 4, F_MUTABLE_BIT);
+        TaskData *taskData = machineDependent->CreateTaskData();
+        taskData->threadObject = (ThreadObject*)alloc(taskData, sizeof(ThreadObject) / sizeof(PolyWord), F_MUTABLE_BIT);
         taskData->threadObject->index = TAGGED(0); // Index 0
         // The initial thread is set to accept broadcast interrupt requests
         // and handle them synchronously.  This is for backwards compatibility.
         taskData->threadObject->flags = TAGGED(PFLAG_BROADCAST|PFLAG_ASYNCH); // Flags
         taskData->threadObject->threadLocal = TAGGED(0); // Empty thread-local store
         taskData->threadObject->requestCopy = TAGGED(0); // Cleared interrupt state
+        taskData->threadObject->mlStackSize = TAGGED(0); // Unlimited stack size
+        for (unsigned i = 0; i < sizeof(taskData->threadObject->debuggerSlots)/sizeof(PolyWord); i++)
+            taskData->threadObject->debuggerSlots[i] = TAGGED(0);
 #if defined(HAVE_WINDOWS_H)
         taskData->threadHandle = mainThreadHandle;
 #endif
@@ -1261,8 +1208,7 @@ void Processes::BeginRootThread(PolyObject *rootFunction)
         if (taskData->stack == 0)
             ::Exit("Unable to create the initial thread - insufficient memory");
 
-        machineDependent->InitStackFrame(taskData, taskData->stack,
-                taskData->saveVec.push(rootFunction), (Handle)0);
+        taskData->InitStackFrame(taskData, taskData->saveVec.push(rootFunction), (Handle)0);
 
         // Create a packet for the Interrupt exception once so that we don't have to
         // allocate when we need to raise it.
@@ -1308,7 +1254,7 @@ void Processes::BeginRootThread(PolyObject *rootFunction)
         if (debugOptions & DEBUG_THREADS)
             Log("THREAD: Forked initial root thread %p\n", taskData);
     }
-    catch (std::bad_alloc a) {
+    catch (std::bad_alloc &) {
         ::Exit("Unable to create the initial thread - insufficient memory");
     }
 
@@ -1322,7 +1268,7 @@ void Processes::BeginRootThread(PolyObject *rootFunction)
         bool signalThreadRunning = false;
         for (unsigned i = 0; i < taskArraySize; i++)
         {
-            ProcessTaskData *p = taskArray[i];
+            TaskData *p = taskArray[i];
             if (p)
             {
                 if (p == sigTask) signalThreadRunning = true;
@@ -1332,8 +1278,7 @@ void Processes::BeginRootThread(PolyObject *rootFunction)
                 {
                     allStopped = false;
                     // It must be running - interrupt it if we are waiting.
-                    if (threadRequest != 0)
-                        machineDependent->InterruptCode(p);
+                    if (threadRequest != 0) p->InterruptCode();
                 }
                 else
                 {
@@ -1380,7 +1325,7 @@ void Processes::BeginRootThread(PolyObject *rootFunction)
             // Set this to kill the threads.
             for (unsigned i = 0; i < taskArraySize; i++)
             {
-                ProcessTaskData *taskData = taskArray[i];
+                TaskData *taskData = taskArray[i];
                 if (taskData && taskData->requests != kRequestKill)
                     MakeRequest(taskData, kRequestKill);
             }
@@ -1408,7 +1353,7 @@ void Processes::BeginRootThread(PolyObject *rootFunction)
         unsigned threadsInML = 0;
         for (unsigned j = 0; j < taskArraySize; j++)
         {
-            ProcessTaskData *taskData = taskArray[j];
+            TaskData *taskData = taskArray[j];
             if (taskData)
             {
                 // This gets the values last time it was in the RTS.
@@ -1432,28 +1377,31 @@ void Processes::BeginRootThread(PolyObject *rootFunction)
         crowbarLock.Signal();
         crowbarStopped.Wait(&shutdownLock);
     }
+    shutdownLock.Unlock(); // So it's unlocked when we delete it
     finish(exitResult); // Close everything down and exit.
 }
 
 // Create a new thread.  Returns the ML thread identifier object if it succeeds.
 // May raise an exception.
-Handle Processes::ForkThread(ProcessTaskData *taskData, Handle threadFunction,
-                           Handle args, PolyWord flags)
+Handle Processes::ForkThread(TaskData *taskData, Handle threadFunction,
+                           Handle args, PolyWord flags, PolyWord stacksize)
 {
     if (singleThreaded)
         raise_exception_string(taskData, EXC_thread, "Threads not available");
 
     try {
         // Create a taskData object for the new thread
-        ProcessTaskData *newTaskData = new ProcessTaskData;
-        newTaskData->mdTaskData = machineDependent->CreateTaskData();
+        TaskData *newTaskData = machineDependent->CreateTaskData();
         // We allocate the thread object in the PARENT's space
-        Handle threadId = alloc_and_save(taskData, 4, F_MUTABLE_BIT);
+        Handle threadId = alloc_and_save(taskData, sizeof(ThreadObject) / sizeof(PolyWord), F_MUTABLE_BIT);
         newTaskData->threadObject = (ThreadObject*)DEREFHANDLE(threadId);
         newTaskData->threadObject->index = TAGGED(0);
         newTaskData->threadObject->flags = flags; // Flags
         newTaskData->threadObject->threadLocal = TAGGED(0); // Empty thread-local store
         newTaskData->threadObject->requestCopy = TAGGED(0); // Cleared interrupt state
+        newTaskData->threadObject->mlStackSize = stacksize;
+        for (unsigned i = 0; i < sizeof(newTaskData->threadObject->debuggerSlots)/sizeof(PolyWord); i++)
+            newTaskData->threadObject->debuggerSlots[i] = TAGGED(0);
 
         unsigned thrdIndex;
         schedLock.Lock();
@@ -1474,8 +1422,8 @@ Handle Processes::ForkThread(ProcessTaskData *taskData, Handle threadFunction,
 
         if (thrdIndex == taskArraySize) // Need to expand the array
         {
-            ProcessTaskData **newArray =
-                (ProcessTaskData **)realloc(taskArray, sizeof(ProcessTaskData *)*(taskArraySize+1));
+            TaskData **newArray =
+                (TaskData **)realloc(taskArray, sizeof(TaskData *)*(taskArraySize+1));
             if (newArray)
             {
                 taskArray = newArray;
@@ -1502,7 +1450,7 @@ Handle Processes::ForkThread(ProcessTaskData *taskData, Handle threadFunction,
 
         // Allocate anything needed for the new stack in the parent's heap.
         // The child still has inMLHeap set so mustn't GC.
-        machineDependent->InitStackFrame(taskData, newTaskData->stack, threadFunction, args);
+        newTaskData->InitStackFrame(taskData, threadFunction, args);
 
         // Now actually fork the thread.
         bool success = false;
@@ -1540,7 +1488,7 @@ Handle Processes::ForkThread(ProcessTaskData *taskData, Handle threadFunction,
 
         raise_exception_string(taskData, EXC_thread, "Thread creation failed");
     }
-    catch (std::bad_alloc a) {
+    catch (std::bad_alloc &) {
             raise_exception_string(taskData, EXC_thread, "Insufficient memory");
     }
 }
@@ -1550,9 +1498,9 @@ Handle Processes::ForkThread(ProcessTaskData *taskData, Handle threadFunction,
 bool Processes::ForkFromRTS(TaskData *taskData, Handle proc, Handle arg)
 {
     try {
-        (void)ForkThread((ProcessTaskData*)taskData, proc, arg, TAGGED(PFLAG_SYNCH));
+        (void)ForkThread(taskData, proc, arg, TAGGED(PFLAG_SYNCH), TAGGED(0));
         return true;
-    } catch (IOException)
+    } catch (IOException &)
     {
         // If it failed
         return false;
@@ -1563,7 +1511,7 @@ bool Processes::ForkFromRTS(TaskData *taskData, Handle proc, Handle arg)
 bool Processes::ProcessAsynchRequests(TaskData *taskData)
 {
     bool wasInterrupted = false;
-    ProcessTaskData *ptaskData = (ProcessTaskData *)taskData;
+    TaskData *ptaskData = taskData;
 
     schedLock.Lock();
 
@@ -1593,7 +1541,7 @@ bool Processes::ProcessAsynchRequests(TaskData *taskData)
                 ptaskData->threadObject->requestCopy = TAGGED(0); // And in the ML copy
                 schedLock.Unlock();
                 // Don't actually throw the exception here.
-                machineDependent->SetException(taskData, interrupt_exn);
+                taskData->SetException(interrupt_exn);
                 wasInterrupted = true;
             }
             else schedLock.Unlock();
@@ -1627,7 +1575,7 @@ bool Processes::ProcessAsynchRequests(TaskData *taskData)
 // called from IO routines which may block.
 void Processes::TestSynchronousRequests(TaskData *taskData)
 {
-    ProcessTaskData *ptaskData = (ProcessTaskData *)taskData;
+    TaskData *ptaskData = taskData;
     schedLock.Lock();
     switch (ptaskData->requests)
     {
@@ -1646,8 +1594,8 @@ void Processes::TestSynchronousRequests(TaskData *taskData)
                 ptaskData->requests = kRequestNone; // Clear this
                 ptaskData->threadObject->requestCopy = TAGGED(0);
                 schedLock.Unlock();
-                machineDependent->SetException(taskData, interrupt_exn);
-                throw IOException(EXC_EXCEPTION);
+                taskData->SetException(interrupt_exn);
+                throw IOException();
             }
             else schedLock.Unlock();
         }
@@ -1665,7 +1613,7 @@ void Processes::TestAnyEvents(TaskData *taskData)
 {
     TestSynchronousRequests(taskData);
     if (ProcessAsynchRequests(taskData))
-        throw IOException(EXC_EXCEPTION);
+        throw IOException();
 }
 
 // Stop.  Usually called by one of the threads but
@@ -1806,7 +1754,7 @@ void Processes::ProfileInterrupt(void)
         {
             for (unsigned i = 0; i < taskArraySize; i++)
             {
-                ProcessTaskData *p = taskArray[i];
+                TaskData *p = taskArray[i];
                 if (p && p->threadHandle)
                 {
                     if (testCPUtime(p->threadHandle, p->lastCPUTime))
@@ -1844,12 +1792,13 @@ void Processes::StartProfiling(void)
 {
 #ifdef HAVE_WINDOWS_H
     DWORD threadId;
+    extern FILE *polyStdout;
     if (profilingHd)
         return;
     ResetEvent(hStopEvent);
     profilingHd = CreateThread(NULL, 0, ProfilingTimer, NULL, 0, &threadId);
     if (profilingHd == NULL)
-        fputs("Creating ProfilingTimer thread failed.\n", stdout); 
+        fputs("Creating ProfilingTimer thread failed.\n", polyStdout);
     /* Give this a higher than normal priority so it pre-empts the main
        thread.  Without this it will tend only to be run when the main
        thread blocks for some reason. */
@@ -1861,10 +1810,10 @@ void Processes::StartProfiling(void)
     // are paused this may not actually be necessary.
     for (unsigned i = 0; i < taskArraySize; i++)
     {
-        ProcessTaskData *taskData = taskArray[i];
+        TaskData *taskData = taskArray[i];
         if (taskData)
         {
-            machineDependent->InterruptCode(taskData);
+            taskData->InterruptCode();
         }
     }
     StartProfilingTimer(); // Start the timer in the root thread.
@@ -1886,7 +1835,7 @@ void Processes::StopProfiling(void)
 // arrives.  There should only be a single thread waiting here.
 bool Processes::WaitForSignal(TaskData *taskData, PLock *sigLock)
 {
-    ProcessTaskData *ptaskData = (ProcessTaskData *)taskData;
+    TaskData *ptaskData = taskData;
     // We need to hold the signal lock until we have acquired schedLock.
     schedLock.Lock();
     sigLock->Unlock();
@@ -1928,7 +1877,7 @@ void Processes::SignalArrived(void)
 // data is cleared.
 static void threaddata_destructor(void *p)
 {
-    ProcessTaskData *pt = (ProcessTaskData *)p;
+    TaskData *pt = (TaskData *)p;
     pt->threadExited = true;
 }
 #endif
@@ -2029,10 +1978,10 @@ void Processes::GarbageCollect(ScanAddress *process)
     }
 }
 
-void ProcessTaskData::GarbageCollect(ScanAddress *process)
+void TaskData::GarbageCollect(ScanAddress *process)
 {
     saveVec.gcScan(process);
-    if (stack != 0) process->ScanAddressesInStack(stack);
+    if (stack != 0) GCStack(process);
 
     if (threadObject != 0)
     {
@@ -2090,13 +2039,14 @@ extern unsigned NumberOfProcessors(void)
 // Return the number of physical processors.  If hyperthreading is
 // enabled this returns less than NumberOfProcessors.  Returns zero if
 // it cannot be determined.
-#if (defined(_WIN32) && defined(HAVE_SYSTEM_LOGICAL_PROCESSOR_INFORMATION))
+// This can be used in Cygwin as well as native Windows.
+#if (defined(HAVE_SYSTEM_LOGICAL_PROCESSOR_INFORMATION))
 typedef BOOL (WINAPI *GETP)(SYSTEM_LOGICAL_PROCESSOR_INFORMATION*, PDWORD);
 
 // Windows - use GetLogicalProcessorInformation if it's available.
 static unsigned WinNumPhysicalProcessors(void)
 {
-    GETP getProcInfo = (GETP) GetProcAddress(GetModuleHandle("kernel32"), "GetLogicalProcessorInformation");
+    GETP getProcInfo = (GETP) GetProcAddress(GetModuleHandle(_T("kernel32")), "GetLogicalProcessorInformation");
     if (getProcInfo == 0) return 0;
 
     // It's there - use it.
@@ -2167,7 +2117,7 @@ static unsigned LinuxNumPhysicalProcessors(void)
 extern unsigned NumberOfPhysicalProcessors(void)
 {
     unsigned numProcs = 0;
-#if (defined(_WIN32) && defined(HAVE_SYSTEM_LOGICAL_PROCESSOR_INFORMATION))
+#if (defined(HAVE_SYSTEM_LOGICAL_PROCESSOR_INFORMATION))
     numProcs = WinNumPhysicalProcessors();
     if (numProcs != 0) return numProcs;
 #endif
