@@ -199,33 +199,6 @@ void ScanAddress::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord
     } while(1);
 }
 
-// Scan addresses in a stack space.
-void ScanAddress::ScanAddressesInStack(StackSpace *stackSpace)
-{
-    StackObject *stack = stackSpace->stack();
-    PolyWord *stackPtr = stack->p_sp; // Save this BEFORE we update
-    PolyWord *stackEnd = stackSpace->top;
-
-    // Either this is TAGGED(0) indicating a retry or it's a code pointer.
-    if (stack->p_pc != TAGGED(0).AsCodePtr())
-        stack->p_pc = ScanStackAddress (PolyWord::FromCodePtr(stack->p_pc), stackSpace, true).AsCodePtr();
-
-    // Stack pointer and handler pointers
-    stack->p_sp =
-        ScanStackAddress (PolyWord::FromStackAddr(stack->p_sp), stackSpace, false).AsStackAddr();
-    stack->p_hr =
-        ScanStackAddress (PolyWord::FromStackAddr(stack->p_hr), stackSpace, false).AsStackAddr();
-
-    // The checked registers.
-    for (POLYUNSIGNED i = 0; i < stack->p_nreg; i++)
-        stack->p_reg[i] = ScanStackAddress(stack->p_reg[i], stackSpace, false);
-
-    // Now the values on the stack.
-    for (PolyWord *q = stackPtr; q < stackEnd; q++)
-        *q = ScanStackAddress(*q, stackSpace, false);
-}
-
-
 void ScanAddress::ScanAddressesInRegion(PolyWord *region, PolyWord *end)
 {
     PolyWord *pt = region;
@@ -564,7 +537,7 @@ void RecursiveScanWithStack::PushToStack(PolyObject *obj)
                     stack->nextStack = s;
                 stack = s;
             }
-            catch (std::bad_alloc a) {
+            catch (std::bad_alloc &) {
                 StackOverflow();
                 return;
             }
