@@ -1,12 +1,11 @@
 /*
     Title:  scanaddrs.h - Scan addresses in objects
 
-    Copyright (c) 2006-8, 2012 David C.J. Matthews
+    Copyright (c) 2006-8, 2012, 2015 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -72,9 +71,6 @@ public:
     // Typically used to scan or update addresses in the mutable area.
     void ScanAddressesInRegion(PolyWord *region, PolyWord *endOfRegion);
 
-    // Scan addresses in a stack space.
-    void ScanAddressesInStack(StackSpace *stackSpace);
-
 protected:
     // General object processor.
     // If the object is a word object calls ScanAddressesAt for all the addresses.
@@ -89,6 +85,7 @@ protected:
 
     void ScanAddressesInObject(PolyObject *base) { ScanAddressesInObject(base, base->LengthWord()); }
 
+public:
     // Process a word found on the stack and return the (updated) value.  Stacks may contain
     // code addresses and addresses within the current stack.  "isCode" is set to true for
     // the program counter which may be a code address but not satisfy IsCodeObject.
@@ -96,7 +93,6 @@ protected:
     // ScanObjectAddress for the base addresses of all other addresses.
     PolyWord ScanStackAddress(PolyWord val, StackSpace *base, bool isCode);
 
-public:
     // Extract a constant from the code.
     static PolyWord GetConstantValue(byte *addressOfConstant, ScanRelocationKind code);
     // Store a constant in the code.
@@ -111,7 +107,7 @@ public:
     virtual void ScanAddressesInObject(PolyObject *base, POLYUNSIGNED lengthWord);
     // Have to redefine this for some reason.
     void ScanAddressesInObject(PolyObject *base)
-        { RecursiveScan::ScanAddressesInObject(base, base->LengthWord()); }
+        { ScanAddressesInObject(base, base->LengthWord()); }
 
 protected:
     // The derived class must provide a stack.
