@@ -653,14 +653,6 @@ void ProcessAddToVector::PushToStack(PolyObject *obj)
     addStack[asp++] = obj;
 }
 
-// There is an optimisation bug in the 64-bit MS C compiler which causes it to convert
-// if (obj->LengthWord() & _OBJ_GC_MARK) obj->SetLengthWord(OBJ_SET_DEPTH(0))
-// into obj->SetLengthWord(obj->LengthWord() & _OBJ_GC_MARK ? OBJ_SET_DEPTH(0) : obj->LengthWord())
-// That will crash if obj points into the permanent immutable area which is read-only.
-#if(defined(_MSC_VER) && defined(_WIN64))
-#pragma optimize( "", off )
-#endif
-
 void ProcessAddToVector::ProcessRoot(PolyObject *root)
 {
     // Mark the initial object
@@ -751,10 +743,6 @@ void ProcessAddToVector::ProcessRoot(PolyObject *root)
     }
 }
 
-#if(defined(_MSC_VER) && defined(_WIN64))
-#pragma optimize( "", off )
-#endif
-
 static void RestoreLengthWords(DepthVector *vec)
 {
    // Restore the length words.
@@ -794,7 +782,7 @@ bool ShareDataClass::RunShareData(PolyObject *root)
         ProcessAddToVector addToVector(this);
         addToVector.ProcessRoot(root);
     }
-    catch (MemoryException)
+    catch (MemoryException &)
     {
         // If we ran out of memory we may still be able to process what we have.
         // That will also do any clean-up.
