@@ -19,10 +19,12 @@
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 */
-#ifdef WIN32
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#elif defined(_WIN32)
 #include "winconfig.h"
 #else
-#include "config.h"
+#error "No configuration file"
 #endif
 
 #ifdef HAVE_STDIO_H
@@ -64,24 +66,18 @@
 #include <stdlib.h> // For malloc
 #endif
 
-#ifdef HAVE_SEMAPHORE_H
+#if ((!defined(_WIN32) || defined(__CYGWIN__)) && defined(HAVE_SEMAPHORE_H))
+// Don't include semaphore.h on Mingw.  It's provided but doesn't compile.
 #include <semaphore.h>
 #endif
 
-#ifdef HAVE_SYS_TYPES_H
-#include <sys/types.h>
+#if ((!defined(_WIN32) || defined(__CYGWIN__)) && defined(HAVE_LIBPTHREAD) && defined(HAVE_PTHREAD_H) && defined(HAVE_SEMAPHORE_H))
+// If we have the pthread library and header and we have semaphores we can use the pthread
+// signalling mechanism.  But if this is a native Windows build we don't use semaphores or
+// pthread even if they're provided.
+#define USE_PTHREAD_SIGNALS 1
 #endif
 
-#ifdef HAVE_SYS_STAT_H
-#include <sys/stat.h>
-#endif
-
-#ifdef HAVE_FCNTL_H
-#include <fcntl.h>
-#endif
-
-
-
 /*
 Signal handling is complicated in a multi-threaded environment.
 The pthread mutex and condition variables are not safe to use in a
@@ -109,7 +105,7 @@ int sigaltstack(const stack_t *, stack_t *);
 #include "scanaddrs.h"
 #include "locking.h"
 
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
 #include "Console.h"
 #endif
 
@@ -133,10 +129,10 @@ unsigned receivedSignalCount = 0; // Incremented each time we get a signal
 // not the "handler" field.
 static PLock sigLock;
 
-#if (defined(HAVE_LIBPTHREAD) && defined(HAVE_PTHREAD_H) && defined(HAVE_SEMAPHORE_H))
-static pthread_t detectionThreadId; // Thread processing signals.
-static sem_t *waitSema;
+#ifdef USE_PTHREAD_SIGNALS
+static PSemaphore *waitSema;
 static int lastSignals[NSIG];
+static bool terminate = false;
 #endif
 
 
@@ -157,7 +153,7 @@ static void signalArrived(int sig)
 void markSignalInuse(int sig)
 {
     sigData[sig].nonMaskable = true;
-#if (defined(HAVE_LIBPTHREAD) && defined(HAVE_PTHREAD_H))
+#ifdef USE_PTHREAD_SIGNALS
     // Enable this signal.
     sigset_t sigset;
     sigemptyset(&sigset);
@@ -174,7 +170,7 @@ static PolyWord findHandler(int sig)
     else return sigData[sig].handler;
 }
 
-#ifdef WINDOWS_PC
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
 // This is called to simulate a SIGINT in Windows.
 void RequestConsoleInterrupt(void)
 {
@@ -185,12 +181,12 @@ void RequestConsoleInterrupt(void)
 }
 #endif
 
-#if (defined(HAVE_LIBPTHREAD) && defined(HAVE_PTHREAD_H) && defined(HAVE_SEMAPHORE_H))
+#ifdef USE_PTHREAD_SIGNALS
 // Request the main thread to change the blocking state of a signal.
 class SignalRequest: public MainThreadRequest
 {
 public:
-    SignalRequest(int s, int r): signl(s), state(r) {}
+    SignalRequest(int s, int r): MainThreadRequest(MTP_SIGHANDLER), signl(s), state(r) {}
 
     virtual void Perform();
     int signl, state;
@@ -203,7 +199,7 @@ static void handle_signal(SIG_HANDLER_ARGS(s, c))
     {
         lastSignals[s]++; // Assume this is atomic with respect to reading.
         // Wake the signal detection thread.
-        sem_post(waitSema);
+        waitSema->Signal();
     }
 }
 
@@ -238,7 +234,7 @@ void SignalRequest::Perform()
    I've used a general dispatch function here to allow for future expansion. */
 Handle Sig_dispatch_c(TaskData *taskData, Handle args, Handle code)
 {
-    int c = get_C_long(taskData, DEREFWORDHANDLE(code));
+    int c = get_C_int(taskData, DEREFWORDHANDLE(code));
     switch (c)
     {
     case 0: /* Set up signal handler. */
@@ -252,10 +248,10 @@ Handle Sig_dispatch_c(TaskData *taskData, Handle args, Handle code)
                 PLocker locker(&sigLock);
                 // We have to pass this to the main thread to 
                 // set up the signal handler.
-                sign = get_C_long(taskData, DEREFHANDLE(args)->Get(0));
+                sign = get_C_int(taskData, DEREFHANDLE(args)->Get(0));
                 /* Decode the action if it is Ignore or Default. */
                 if (IS_INT(DEREFHANDLE(args)->Get(1)))
-                    action = UNTAGGED(DEREFHANDLE(args)->Get(1));
+                    action = (int)UNTAGGED(DEREFHANDLE(args)->Get(1));
                 else action = HANDLE_SIG; /* Set the handler. */
                 if (sign <= 0 || sign >= NSIG)
                     raise_syscall(taskData, "Invalid signal value", EINVAL);
@@ -271,7 +267,7 @@ Handle Sig_dispatch_c(TaskData *taskData, Handle args, Handle code)
             // we affect is SIGINT and that is handled by RequestConsoleInterrupt.
             if (! sigData[sign].nonMaskable)
             {
-#if (defined(HAVE_LIBPTHREAD) && defined(HAVE_PTHREAD_H) && defined(HAVE_SEMAPHORE_H))
+#ifdef USE_PTHREAD_SIGNALS
                 SignalRequest request(sign, action);
                 processes->MakeRootRequest(taskData, &request);
 #endif
@@ -327,7 +323,7 @@ Handle Sig_dispatch_c(TaskData *taskData, Handle args, Handle code)
             char msg[100];
             sprintf(msg, "Unknown signal function: %d", c);
             raise_exception_string(taskData, EXC_Fail, msg);
-			return 0;
+            return 0;
         }
     }
 }
@@ -336,7 +332,7 @@ Handle Sig_dispatch_c(TaskData *taskData, Handle args, Handle code)
 // This is really only needed for profiling timer signals.
 void initThreadSignals(TaskData *taskData)
 {
-#if (!(defined(WINDOWS_PC)||defined(MACOSX)))
+#if (!(defined(_WIN32)||defined(MACOSX)))
     // On the i386, at least, we need to set up a signal stack for
     // each thread if it might receive a signal.  ML code checks for
     // stack overflow but a signal could result in C code being
@@ -368,7 +364,7 @@ void initThreadSignals(TaskData *taskData)
     ASSERT(sigaltstack_result == 0);
 #endif
 #endif /* not the PC */
-#if (defined(HAVE_LIBPTHREAD) && defined(HAVE_PTHREAD_H))
+#ifdef USE_PTHREAD_SIGNALS
     // Block all signals except those marked as in use by the RTS so
     // that they will only be picked up by the signal detection thread.
     // Since the signal mask is inherited we really don't need to do
@@ -386,7 +382,7 @@ void initThreadSignals(TaskData *taskData)
 
 
 /* General purpose function to set up a signal handler. */
-#ifndef WINDOWS_PC
+#if (!defined(_WIN32) || defined(__CYGWIN__))
 
 bool setSignalHandler(int sig, signal_handler_type func)
 {
@@ -456,13 +452,20 @@ class SigHandler: public RtsModule
 {
 public:
     virtual void Init(void);
+    virtual void Stop(void);
     virtual void GarbageCollect(ScanAddress * /*process*/);
+
+#ifdef USE_PTHREAD_SIGNALS
+    SigHandler() { threadRunning = false; }
+    pthread_t detectionThreadId;
+    bool      threadRunning;
+#endif
 };
 
 // Declare this.  It will be automatically added to the table.
 static SigHandler sighandlerModule;
 
-#if (defined(HAVE_LIBPTHREAD) && defined(HAVE_PTHREAD_H) && defined(HAVE_SEMAPHORE_H))
+#ifdef USE_PTHREAD_SIGNALS
 // This thread is really only to convert between POSIX semaphores and
 // pthread condition variables.  It waits for a semphore to be released by the
 // signal handler running on the main thread and then wakes up the ML handler
@@ -485,11 +488,8 @@ static void *SignalDetectionThread(void *)
         // Wait until we are woken up by an arriving signal.
         // waitSema will be incremented for each signal so we should
         // not block until we have processed them all.
-        while (sem_wait(waitSema) == -1)
-        {
-            if (errno != EINTR)
-                return 0;
-        }
+        if (! waitSema->Wait() || terminate) return 0;
+
         for (int j = 1; j < NSIG; j++)
         {
             if (readSignals[j] < lastSignals[j])
@@ -502,32 +502,6 @@ static void *SignalDetectionThread(void *)
 }
 #endif
 
-#if (defined(HAVE_SEMAPHORE_H))
-static sem_t waitSemaphore;
-
-// Initialise a semphore.  Tries to create an unnamed semaphore if
-// it can but tries a named semaphore if it can't.  Mac OS X only
-// supports named semaphores.
-static sem_t *init_semaphore(sem_t *sema, int init)
-{
-    if (sem_init(sema, 0, init) == 0)
-        return sema;
-#if (defined(__CYGWIN__))
-    // Cygwin doesn't define sem_unlink but that doesn't matter
-    // since sem_init works.
-    return 0;
-#else
-    char semname[30];
-    static int count=0;
-    sprintf(semname, "poly%0d-%0d", (int)getpid(), count++);
-    sema = sem_open(semname, O_CREAT|O_EXCL, 00666, init);
-    if (sema == (sem_t*)SEM_FAILED) return 0;
-    sem_unlink(semname);
-    return sema;
-#endif
-}
-#endif
-
 void SigHandler::Init(void)
 {
     // Mark certain signals as non-maskable since they really
@@ -541,15 +515,15 @@ void SigHandler::Init(void)
 #ifdef SIGILL
     sigData[SIGILL].nonMaskable = true;
 #endif
-#if (defined(HAVE_LIBPTHREAD) && defined(HAVE_PTHREAD_H) && defined(HAVE_SEMAPHORE_H))
+#ifdef USE_PTHREAD_SIGNALS
+    static PSemaphore waitSemaphore;
     // Initialise the "wait" semaphore so that it blocks immediately.
-    waitSema = init_semaphore(&waitSemaphore, 0);
-    if (waitSema == 0) return;
+    if (! waitSemaphore.Init(0, NSIG)) return;
+    waitSema = &waitSemaphore;
     // Create a new thread to handle signals synchronously.
     // for it to finish.
     pthread_attr_t attrs;
     pthread_attr_init(&attrs);
-    pthread_attr_setdetachstate(&attrs, PTHREAD_CREATE_DETACHED);
 #ifdef PTHREAD_STACK_MIN
 #if (PTHREAD_STACK_MIN < 4096)
     pthread_attr_setstacksize(&attrs, 4096); // But not too small: FreeBSD makes it 2k
@@ -557,20 +531,27 @@ void SigHandler::Init(void)
     pthread_attr_setstacksize(&attrs, PTHREAD_STACK_MIN); // Only small stack.
 #endif
 #endif
-    pthread_create(&detectionThreadId, &attrs, SignalDetectionThread, 0);
+    threadRunning = pthread_create(&detectionThreadId, &attrs, SignalDetectionThread, 0) == 0;
     pthread_attr_destroy(&attrs);
 #endif
 }
 
+// Wait for the signal thread to finish before the semaphore is deleted in the
+// final clean-up.  Failing to do this causes a hang in Mac OS X.
+void SigHandler::Stop(void)
+{
+#ifdef USE_PTHREAD_SIGNALS
+    terminate = true;
+    waitSema->Signal();
+    pthread_join(detectionThreadId, NULL);
+#endif
+}
+
 void SigHandler::GarbageCollect(ScanAddress *process)
 {
     for (unsigned i = 0; i < NSIG; i++)
     {
-        if (sigData[i].handler != PolyWord::FromUnsigned(0) && sigData[i].handler.IsDataPtr())
-        {
-            PolyObject *obj = sigData[i].handler.AsObjPtr();
-            process->ScanRuntimeAddress(&obj, ScanAddress::STRENGTH_STRONG);
-            sigData[i].handler = obj;
-        }
+        if (sigData[i].handler != PolyWord::FromUnsigned(0))
+            process->ScanRuntimeWord(&sigData[i].handler);
     }
 }
