@@ -1,12 +1,11 @@
 /*
     Title:      Operating Specific functions: Windows version.
 
-    Copyright (c) 2000 David C. J. Matthews
+    Copyright (c) 2000, 2015 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -26,17 +25,13 @@
 #error "No configuration file"
 #endif
 
+#include <winsock2.h>
+#include <windows.h>
+
 #ifdef HAVE_STDIO_H
 #include <stdio.h>
 #endif
 
-#include <windows.h>
-#ifdef USEWINSOCK2
-#include <winsock2.h>
-#else
-#include <winsock.h>
-#endif
-
 #ifdef HAVE_TCHAR_H
 #include <tchar.h>
 #endif
@@ -186,12 +181,12 @@ static Handle make_handle_entry(TaskData *taskData)
             {
                 POLYUNSIGNED oldMax = maxHandleTab;
                 maxHandleTab += maxHandleTab/2;
-                handleTable =
-                    (PHANDLETAB)realloc(handleTable,
-                                    maxHandleTab*sizeof(HANDLETAB));
+                void *p = realloc(handleTable, maxHandleTab*sizeof(HANDLETAB));
+                // If there's insufficient memory leave the old table.
+                if (p == 0) raise_syscall(taskData, "Insufficient memory", ENOMEM);
+                handleTable = (PHANDLETAB)p;
                 /* Clear the new space. */
-                memset(handleTable+oldMax, 0,
-                        (maxHandleTab-oldMax)*sizeof(HANDLETAB));
+                memset(handleTable+oldMax, 0, (maxHandleTab-oldMax)*sizeof(HANDLETAB));
             }
         }
     } while (handle_no >= maxHandleTab);
@@ -298,7 +293,7 @@ HKEY hkPredefinedKeyTab[] =
 
 Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 {
-    int c = get_C_int(taskData, DEREFWORD(code));
+    unsigned c = get_C_unsigned(taskData, DEREFWORD(code));
     switch (c)
     {
     case 0: /* Return our OS type.  Not in any structure. */
@@ -540,31 +535,27 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
         {
             FILETIME ftUTC, ftLocal;
             /* Get the file time. */
-            get_C_pair(taskData, DEREFWORDHANDLE(args),
-                    &ftUTC.dwHighDateTime, &ftUTC.dwLowDateTime);
+            getFileTimeFromArb(taskData, args, &ftUTC);
             if (! FileTimeToLocalFileTime(&ftUTC, &ftLocal))
                 raise_syscall(taskData, "FileTimeToLocalFileTime failed",
                         -(int)GetLastError());
-            return Make_arb_from_pair(taskData, ftLocal.dwHighDateTime,
-                        ftLocal.dwLowDateTime);
+            return Make_arb_from_Filetime(taskData, ftLocal);
         }
 
     case 1031: // Convert local time values to UTC. -- No longer used??
         {
             FILETIME ftUTC, ftLocal;
             /* Get the file time. */
-            get_C_pair(taskData, DEREFWORDHANDLE(args),
-                    &ftLocal.dwHighDateTime, &ftLocal.dwLowDateTime);
+            getFileTimeFromArb(taskData, args, &ftLocal);
             if (! LocalFileTimeToFileTime(&ftLocal, &ftUTC))
                 raise_syscall(taskData, "LocalFileTimeToFileTime failed",
                         -(int)GetLastError());
-            return Make_arb_from_pair(taskData, ftUTC.dwHighDateTime,
-                        ftUTC.dwLowDateTime);
+            return Make_arb_from_Filetime(taskData, ftUTC);
         }
 
     case 1032: // Get volume information.
         {
-            char rootName[MAX_PATH], volName[MAX_PATH], sysName[MAX_PATH];
+            TCHAR rootName[MAX_PATH], volName[MAX_PATH], sysName[MAX_PATH];
             DWORD dwVolSerial, dwMaxComponentLen, dwFlags;
             Handle volHandle, sysHandle, serialHandle, maxCompHandle;
             Handle resultHandle;
@@ -591,7 +582,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1033:
         {
-            char fileName[MAX_PATH], execName[MAX_PATH];
+            TCHAR fileName[MAX_PATH], execName[MAX_PATH];
             POLYUNSIGNED length = Poly_string_to_C(DEREFWORD(args), fileName, MAX_PATH);
             HINSTANCE hInst;
             if (length > MAX_PATH)
@@ -650,8 +641,8 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
             Handle handToken;
             PHANDLETAB pTab;
             HCONV hcDDEConv;
-            char *serviceName = Poly_string_to_C_alloc(args->WordP()->Get(0));
-            char *topicName = Poly_string_to_C_alloc(args->WordP()->Get(1));
+            TCHAR *serviceName = Poly_string_to_T_alloc(args->WordP()->Get(0));
+            TCHAR *topicName = Poly_string_to_T_alloc(args->WordP()->Get(1));
             /* Send a request to the main thread to do the work. */
             hcDDEConv = StartDDEConversation(serviceName, topicName);
             free(serviceName); free(topicName);
@@ -694,16 +685,17 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
     case 1050: // Get version data
         {
             OSVERSIONINFO osver;
-            Handle resVal, major, minor, build, platform, version;
+            ZeroMemory(&osver, sizeof(OSVERSIONINFO));
             osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
+            // GetVersionEx is deprecated in Windows 8.1
             if (! GetVersionEx(&osver))
                 raise_syscall(taskData, "GetVersionEx failed", -(int)GetLastError());
-            major = Make_arbitrary_precision(taskData, osver.dwMajorVersion);
-            minor = Make_arbitrary_precision(taskData, osver.dwMinorVersion);
-            build = Make_arbitrary_precision(taskData, osver.dwBuildNumber);
-            platform = Make_arbitrary_precision(taskData, osver.dwPlatformId);
-            version = SAVE(C_string_to_Poly(taskData, osver.szCSDVersion));
-            resVal = alloc_and_save(taskData, 5);
+            Handle major = Make_arbitrary_precision(taskData, osver.dwMajorVersion);
+            Handle minor = Make_arbitrary_precision(taskData, osver.dwMinorVersion);
+            Handle build = Make_arbitrary_precision(taskData, osver.dwBuildNumber);
+            Handle platform = Make_arbitrary_precision(taskData, osver.dwPlatformId);
+            Handle version = SAVE(C_string_to_Poly(taskData, osver.szCSDVersion));
+            Handle resVal = alloc_and_save(taskData, 5);
             DEREFHANDLE(resVal)->Set(0, DEREFWORDHANDLE(major));
             DEREFHANDLE(resVal)->Set(1, DEREFWORDHANDLE(minor));
             DEREFHANDLE(resVal)->Set(2, DEREFWORDHANDLE(build));
@@ -714,23 +706,23 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1051: // Get windows directory
         {
-            char path[MAX_PATH+1];
-            if (GetWindowsDirectory(path, sizeof(path)/sizeof(char)) == 0)
+            TCHAR path[MAX_PATH+1];
+            if (GetWindowsDirectory(path, sizeof(path)/sizeof(TCHAR)) == 0)
                 raise_syscall(taskData, "GetWindowsDirectory failed", -(int)GetLastError());
             return SAVE(C_string_to_Poly(taskData, path));
         }
 
     case 1052: // Get system directory
         {
-            char path[MAX_PATH+1];
-            if (GetSystemDirectory(path, sizeof(path)/sizeof(char)) == 0)
+            TCHAR path[MAX_PATH+1];
+            if (GetSystemDirectory(path, sizeof(path)/sizeof(TCHAR)) == 0)
                 raise_syscall(taskData, "GetSystemDirectory failed", -(int)GetLastError());
             return SAVE(C_string_to_Poly(taskData, path));
         }
 
     case 1053: // Get computer name
         {
-            char name[MAX_COMPUTERNAME_LENGTH +1];
+            TCHAR name[MAX_COMPUTERNAME_LENGTH +1];
             DWORD dwSize = MAX_COMPUTERNAME_LENGTH +1;
             if (GetComputerName(name, &dwSize) == 0)
                 raise_syscall(taskData, "GetComputerName failed", -(int)GetLastError());
@@ -739,7 +731,7 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1054: // Get user name
         {
-            char name[UNLEN +1];
+            TCHAR name[UNLEN +1];
             DWORD dwSize = UNLEN +1;
             if (GetUserName(name, &dwSize) == 0)
                 raise_syscall(taskData, "GetUserName failed", -(int)GetLastError());
@@ -754,11 +746,11 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
 
     case 1101: // Wait for a message.
         {
+            HWND hwnd = *(HWND*)(DEREFWORDHANDLE(args)->Get(0).AsCodePtr());
+            UINT wMsgFilterMin = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(1));
+            UINT wMsgFilterMax = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(2));
             while (1)
             {
-                HWND hwnd = (HWND)get_C_long(taskData, DEREFWORDHANDLE(args)->Get(0)); /* Handles are treated as SIGNED. */
-                UINT wMsgFilterMin = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(1));
-                UINT wMsgFilterMax = get_C_unsigned(taskData, DEREFWORDHANDLE(args)->Get(2));
                 MSG msg;
                 processes->ThreadReleaseMLMemory(taskData);
                 // N.B.  PeekMessage may directly call the window proc resulting in a
@@ -774,10 +766,18 @@ Handle OS_spec_dispatch_c(TaskData *taskData, Handle args, Handle code)
     // case 1102: // Return the address of the window callback function.
 
     case 1103: // Return the application instance.
-        return Make_arbitrary_precision(taskData, (POLYUNSIGNED)hApplicationInstance);
+        {
+            Handle result = alloc_and_save(taskData, 1, F_BYTE_OBJ);
+            *(HINSTANCE*)(result->Word().AsCodePtr()) = hApplicationInstance;
+            return result;
+        }
 
     case 1104: // Return the main window handle
-        return Make_arbitrary_precision(taskData, (POLYUNSIGNED)hMainWindow);
+        {
+            Handle result = alloc_and_save(taskData, 1, F_BYTE_OBJ);
+            *(HWND*)(result->Word().AsCodePtr()) = hMainWindow;
+            return result;
+        }
 
 //    case 1105: // Set the callback function
 
@@ -810,7 +810,6 @@ interpose a thread which can signal an event when input is available.
 */
 static Handle execute(TaskData *taskData, Handle args)
 {
-    char *commandName = 0, *arguments = 0;
     LPCSTR lpszError = "";
     HANDLE hWriteToChild = INVALID_HANDLE_VALUE,
            hReadFromParent = INVALID_HANDLE_VALUE,
@@ -823,8 +822,8 @@ static Handle execute(TaskData *taskData, Handle args)
     Handle handToken = make_handle_entry(taskData);
     PHANDLETAB pTab = &handleTable[STREAMID(handToken)];
 
-    commandName = Poly_string_to_C_alloc(args->WordP()->Get(0));
-    arguments = Poly_string_to_C_alloc(args->WordP()->Get(1));
+    LPTSTR commandName = Poly_string_to_T_alloc(args->WordP()->Get(0));
+    LPTSTR arguments = Poly_string_to_T_alloc(args->WordP()->Get(1));
 
     // Create pipes for connection. Setting the security argument to NULL creates
     // the pipe handles as non-inheritable.  We have to make sure that the
@@ -876,9 +875,9 @@ static Handle execute(TaskData *taskData, Handle args)
     // What should we do about the stderr?  For the moment, inherit the original.
     startupInfo.hStdError = GetStdHandle(STD_ERROR_HANDLE);
 
-    // Try starting the process first using the given name.
-    if (!CreateProcess(commandName,
-            arguments, // Command line
+    // Treat the empty string as NULL.  This is non-standard.
+    if (!CreateProcess(commandName[0] == 0 ? NULL: commandName,
+            arguments[0] == 0 ? NULL: arguments, // Command line
             NULL, NULL, TRUE, // Security attributes. Inherit handles
             CREATE_NO_WINDOW, // creation flags
             NULL, NULL, // Inherit our environment and directory
@@ -947,9 +946,9 @@ static Handle simpleExecute(TaskData *taskData, Handle args)
     startupInfo.hStdError = hNull;
     STARTUPINFO *start = &startupInfo;
 
-    // Try starting the process first using the given name.
-    if (!CreateProcess(commandName,
-            arguments, // Command line
+    // Treat the empty string as NULL.  This is non-standard.
+    if (!CreateProcess(commandName[0] == 0 ? NULL : commandName,
+            arguments[0] == 0 ? NULL : arguments, // Command line
             NULL, NULL, // Security attributes
             TRUE, CREATE_NO_WINDOW, // Inherit handles, creation flags
             NULL, NULL, // Inherit our environment and directory
@@ -1161,11 +1160,17 @@ static Handle queryRegistryKey(TaskData *taskData, Handle args, HKEY hkey)
     // try reading directly into ML store to save copying but
     // it hardly seems worthwhile.
     // Note: It seems that valSize can be zero for some items.
-    if (valSize == 0) resVal = SAVE(Buffer_to_Poly(taskData, "", 0));
+    if (valSize == 0) resVal = SAVE(C_string_to_Poly(taskData, "", 0));
     else
     {
         do {
-            keyValue = (byte*)realloc(keyValue, valSize);
+            byte *newAlloc = (byte*)realloc(keyValue, valSize);
+            if (newAlloc == 0)
+            {
+                free(keyValue);
+                raise_syscall(taskData, "Insufficient memory", ENOMEM);
+            }
+            keyValue = newAlloc;
             lRes = RegQueryValueEx(hkey, valName, 0, &dwType, keyValue, &valSize);
             // In the special case of HKEY_PERFORMANCE_DATA we may need to keep
             // growing the buffer.
@@ -1177,7 +1182,10 @@ static Handle queryRegistryKey(TaskData *taskData, Handle args, HKEY hkey)
             free(keyValue);
             raise_syscall(taskData, "RegQueryValue failed", -lRes);
         }
-        resVal = SAVE(Buffer_to_Poly(taskData, (char*)keyValue, valSize));
+        // If we have a string we have to convert this to ANSI/utf-8.
+        if (dwType == REG_SZ || dwType == REG_MULTI_SZ || dwType == REG_EXPAND_SZ)
+            resVal = SAVE(C_string_to_Poly(taskData, (TCHAR*)keyValue, valSize / sizeof(TCHAR)));
+        else resVal = SAVE(C_string_to_Poly(taskData, (char*)keyValue, valSize));
         free(keyValue);
     }
 
