@@ -1,12 +1,11 @@
 /*
     Title:  polyexports.h 
 
-    Copyright (c) 2006, 2011 David C.J. Matthews
+    Copyright (c) 2006, 2011, 2015 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -103,12 +102,11 @@ extern "C" {
 # endif
 
 extern POLYLIB_API int PolyWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
-                    LPTSTR lpCmdLine, int nCmdShow, exportDescription *exports);
+                    LPSTR lpCmdLine, int nCmdShow, exportDescription *exports);
+#else
+int polymain(int argc, char *argv[], exportDescription *exports);
 #endif
 
-    int polymain(int argc, char *argv[], exportDescription *exports);
-
-
 #ifdef __cplusplus
 };
 #endif
