@@ -89,6 +89,10 @@
 
 /* Define to 1 if you have the <fenv.h> header file. */
 #undef HAVE_FENV_H
+#if (defined(_MSC_VER) && (_MSC_VER >= 1800))
+// Defined in VS 2013
+#define HAVE_FENV_H 1
+#endif
 
 /* Define to 1 if you have the <float.h> header file. */
 #undef HAVE_FLOAT_H
@@ -151,6 +155,7 @@
 
 /* Define to 1 if you have the <inttypes.h> header file. */
 #undef HAVE_INTTYPES_H
+// This was present in VS 2013 but not 2015.
 
 /* Define to 1 if you have the <io.h> header file. */
 #undef HAVE_IO_H
@@ -221,6 +226,7 @@
 
 /* Define to 1 if the system has the type `long long'. */
 #undef HAVE_LONG_LONG
+#define HAVE_LONG_LONG 1
 
 /* Define to 1 if `lstat' has the bug that it succeeds when given the
    zero-length file name argument. */
@@ -377,6 +383,7 @@
 
 /* Define to 1 if you have the <stdint.h> header file. */
 #undef HAVE_STDINT_H
+#define HAVE_STDINT_H 1
 
 /* Define to 1 if you have the <stdio.h> header file. */
 #undef HAVE_STDIO_H
@@ -441,7 +448,7 @@
 
 /* Define to 1 if you have the `sysctlbyname' function. */
 #undef HAVE_SYSCTLBYNAME
- 
+
 /* Define to 1 if the system has the type
    `SYSTEM_LOGICAL_PROCESSOR_INFORMATION'. */
 #undef HAVE_SYSTEM_LOGICAL_PROCESSOR_INFORMATION
@@ -586,22 +593,31 @@
 /* Define to 1 if the system has the type `_Bool'. */
 #undef HAVE__BOOL
 
-/* Define if the host is an ARM (32 bit) */
+/* Define if the host is an ARM (64-bit) */
+/*#undef HOSTARCHITECTURE_AARCH64*/
+
+/* Define if the host is an ARM (32-bit) */
 /*#undef HOSTARCHITECTURE_ARM*/
 
-/* Define if the host is Itanium */
+/* Define if the host is an Itanium */
 /*#undef HOSTARCHITECTURE_IA64*/
 
-/* Define if the host is a Power PC (32 bit) */
+/* Define if the host is a MIPS (32-bit) */
+/*#undef HOSTARCHITECTURE_MIPS*/
+
+/* Define if the host is a PowerPC (32-bit) */
 /*#undef HOSTARCHITECTURE_PPC*/
 
-/* Define if the host is a Sparc (32 bit) */
+/* Define if the host is a PowerPC (64-bit) */
+/*#undef HOSTARCHITECTURE_PPC64*/
+
+/* Define if the host is a Sparc (32-bit) */
 /*#undef HOSTARCHITECTURE_SPARC*/
 
-/* Define if the host is an X86 (32 bit) */
+/* Define if the host is an X86 (32-bit) */
 /*#undef HOSTARCHITECTURE_X86*/
 
-/* Define if the host is an X86 (64 bit) */
+/* Define if the host is an X86 (64-bit) */
 /*#undef HOSTARCHITECTURE_X86_64*/
 
 /* Define to 1 if `lstat' dereferences a symlink specified with a trailing
@@ -719,6 +735,15 @@
 /* Define to `unsigned int' if <sys/types.h> does not define. */
 #undef size_t
 
+/* Define to `int' if <sys/types.h> does not define. */
+// There is an SSIZE_T
+#undef ssize_t
+
+#if defined(_MSC_VER)
+#include <BaseTsd.h>
+typedef SSIZE_T ssize_t;
+#endif
+
 /* Define to `int' if <sys/types.h> doesn't define. */
 #define uid_t int
 
